# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from openstack import exceptions
from openstack import proxy
from openstack import resource
from openstack.shared_file_system.v2 import (
    availability_zone as _availability_zone,
)
from openstack.shared_file_system.v2 import limit as _limit
from openstack.shared_file_system.v2 import quota_class_set as _quota_class_set
from openstack.shared_file_system.v2 import resource_locks as _resource_locks
from openstack.shared_file_system.v2 import share as _share
from openstack.shared_file_system.v2 import share_group as _share_group
from openstack.shared_file_system.v2 import (
    share_group_snapshot as _share_group_snapshot,
)
from openstack.shared_file_system.v2 import (
    share_access_rule as _share_access_rule,
)
from openstack.shared_file_system.v2 import (
    share_export_locations as _share_export_locations,
)
from openstack.shared_file_system.v2 import share_instance as _share_instance
from openstack.shared_file_system.v2 import share_network as _share_network
from openstack.shared_file_system.v2 import (
    share_network_subnet as _share_network_subnet,
)
from openstack.shared_file_system.v2 import share_snapshot as _share_snapshot
from openstack.shared_file_system.v2 import (
    share_snapshot_instance as _share_snapshot_instance,
)
from openstack.shared_file_system.v2 import storage_pool as _storage_pool
from openstack.shared_file_system.v2 import user_message as _user_message


class Proxy(proxy.Proxy):
    _resource_registry = {
        "availability_zone": _availability_zone.AvailabilityZone,
        "share_snapshot": _share_snapshot.ShareSnapshot,
        "storage_pool": _storage_pool.StoragePool,
        "user_message": _user_message.UserMessage,
        "limit": _limit.Limit,
        "share": _share.Share,
        "share_network": _share_network.ShareNetwork,
        "share_network_subnet": _share_network_subnet.ShareNetworkSubnet,
        "share_snapshot_instance": _share_snapshot_instance.ShareSnapshotInstance,  # noqa: E501
        "share_instance": _share_instance.ShareInstance,
        "share_export_locations": _share_export_locations.ShareExportLocation,
        "share_access_rule": _share_access_rule.ShareAccessRule,
        "share_group": _share_group.ShareGroup,
        "share_group_snapshot": _share_group_snapshot.ShareGroupSnapshot,
        "resource_locks": _resource_locks.ResourceLock,
        "quota_class_set": _quota_class_set.QuotaClassSet,
    }

    def availability_zones(self):
        """Retrieve shared file system availability zones

        :returns: A generator of availability zone resources
        :rtype:
            :class:`~openstack.shared_file_system.v2.availability_zone.AvailabilityZone`
        """
        return self._list(_availability_zone.AvailabilityZone)

    def shares(self, details=True, **query):
        """Lists all shares with details

        :param kwargs query: Optional query parameters to be sent to limit
            the shares being returned.  Available parameters include:

            * status: Filters by a share status
            * share_server_id: The UUID of the share server.
            * metadata: One or more metadata key and value pairs as a url
              encoded dictionary of strings.
            * extra_specs: The extra specifications as a set of one or more
              key-value pairs.
            * share_type_id: The UUID of a share type to query resources by.
            * name: The user defined name of the resource to filter resources
              by.
            * snapshot_id: The UUID of the share's base snapshot to filter
              the request based on.
            * host: The host name of the resource to query with.
            * share_network_id: The UUID of the share network to filter
              resources by.
            * project_id: The ID of the project that owns the resource.
            * is_public: A boolean query parameter that, when set to true,
              allows retrieving public resources that belong to
              all projects.
            * share_group_id: The UUID of a share group to filter resource.
            * export_location_id: The export location UUID that can be used
              to filter shares or share instances.
            * export_location_path: The export location path that can be used
              to filter shares or share instances.
            * name~: The name pattern that can be used to filter shares, share
              snapshots, share networks or share groups.
            * description~: The description pattern that can be used to filter
              shares, share snapshots, share networks or share groups.
            * with_count: Whether to show count in API response or not,
              default is False.
            * limit: The maximum number of shares to return.
            * offset: The offset to define start point of share or share group
              listing.
            * sort_key: The key to sort a list of shares.
            * sort_dir: The direction to sort a list of shares. A valid value
              is asc, or desc.

        :returns: Details of shares resources
        :rtype: :class:`~openstack.shared_file_system.v2.share.Share`
        """
        base_path = '/shares/detail' if details else None
        return self._list(_share.Share, base_path=base_path, **query)

    def find_share(self, name_or_id, ignore_missing=True, **query):
        """Find a single share

        :param name_or_id: The name or ID of a share.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :param dict query: Any additional parameters to be passed into
            underlying methods. such as query filters.

        :returns: One :class:`~openstack.shared_file_system.v2.share.Share`
                  or None
        """

        return self._find(
            _share.Share, name_or_id, ignore_missing=ignore_missing, **query
        )

    def get_share(self, share_id):
        """Lists details of a single share

        :param share: The ID of the share to get
        :returns: Details of the identified share
        :rtype: :class:`~openstack.shared_file_system.v2.share.Share`
        """
        return self._get(_share.Share, share_id)

    def delete_share(self, share, ignore_missing=True):
        """Deletes a single share

        :param share: The ID of the share to delete
        :returns: Result of the ``delete``
        :rtype: ``None``
        """
        self._delete(_share.Share, share, ignore_missing=ignore_missing)

    def update_share(self, share_id, **attrs):
        """Updates details of a single share.

        :param share: The ID of the share to update
        :param dict attrs: The attributes to update on the share
        :returns: the updated share
        :rtype: :class:`~openstack.shared_file_system.v2.share.Share`
        """
        return self._update(_share.Share, share_id, **attrs)

    def create_share(self, **attrs):
        """Creates a share from attributes

        :returns: Details of the new share
        :param dict attrs: Attributes which will be used to create
            a :class:`~openstack.shared_file_system.v2.shares.Shares`,
            comprised of the properties on the Shares class. 'size' and 'share'
            are required to create a share.
        :rtype: :class:`~openstack.shared_file_system.v2.share.Share`
        """
        return self._create(_share.Share, **attrs)

    def revert_share_to_snapshot(self, share_id, snapshot_id):
        """Reverts a share to the specified snapshot, which must be
            the most recent one known to manila.

        :param share_id: The ID of the share to revert
        :param snapshot_id: The ID of the snapshot to revert to
        :returns: Result of the ``revert``
        :rtype: ``None``
        """
        res = self._get(_share.Share, share_id)
        res.revert_to_snapshot(self, snapshot_id)

    def manage_share(self, protocol, export_path, service_host, **params):
        """Manage a share.

        :param str protocol: The shared file systems protocol of this share.
        :param str export_path: The export path formatted according to the
            protocol.
        :param str service_host: The manage-share service host.
        :param kwargs params: Optional parameters to be sent. Available
            parameters include:
            * name: The user defined name of the resource.
            * share_type: The name or ID of the share type to be used to create
            the resource.
            * driver_options: A set of one or more key and value pairs, as a
            dictionary of strings, that describe driver options.
            * is_public: The level of visibility for the share.
            * description: The user defiend description of the resource.
            * share_server_id: The UUID of the share server.

        :returns: The share that was managed.
        """

        share = _share.Share()
        return share.manage(
            self, protocol, export_path, service_host, **params
        )

    def unmanage_share(self, share_id):
        """Unmanage the share with the given share ID.

        :param share_id: The ID of the share to unmanage.
        :returns: ``None``
        """

        share_to_unmanage = self._get(_share.Share, share_id)
        share_to_unmanage.unmanage(self)

    def resize_share(
        self, share_id, new_size, no_shrink=False, no_extend=False, force=False
    ):
        """Resizes a share, extending/shrinking the share as needed.

        :param share_id: The ID of the share to resize
        :param new_size: The new size of the share in GiBs. If new_size is
            the same as the current size, then nothing is done.
        :param bool no_shrink: If set to True, the given share is not shrunk,
            even if shrinking the share is required to get the share to the
            given size. This could be useful for extending shares to a minimum
            size, while not shrinking shares to the given size. This defaults
            to False.
        :param bool no_extend: If set to True, the given share is not
            extended, even if extending the share is required to get the share
            to the given size. This could be useful for shrinking shares to a
            maximum size, while not extending smaller shares to that maximum
            size. This defaults to False.
        :param bool force: Whether or not force should be used,
            in the case where the share should be extended.
        :returns: ``None``
        """

        res = self._get(_share.Share, share_id)

        if new_size > res.size and no_extend is not True:
            res.extend_share(self, new_size, force)
        elif new_size < res.size and no_shrink is not True:
            res.shrink_share(self, new_size)

    def share_groups(self, **query):
        """Lists all share groups.

        :param kwargs query: Optional query parameters to be sent to limit
            the share groups being returned.  Available parameters include:

            * status: Filters by a share group status.
            * name: The user defined name of the resource to filter resources
                by.
            * description: The user defined description text that can be used
                to filter resources.
            * project_id: The project ID of the user or service.
            * share_server_id: The UUID of the share server.
            * snapshot_id: The UUID of the share's base snapshot to filter
                the request based on.
            * host: The host name for the back end.
            * share_network_id: The UUID of the share network to filter
                resources by.
            * share_group_type_id: The share group type ID to filter
                share groups.
            * share_group_snapshot_id: The source share group snapshot ID to
                list the share group.
            * share_types: A list of one or more share type IDs. Allows
                filtering share groups.
            * limit: The maximum number of share groups members to return.
            * offset: The offset to define start point of share or share
                group listing.
            * sort_key: The key to sort a list of shares.
            * sort_dir: The direction to sort a list of shares
            * name~: The name pattern that can be used to filter shares,
                share snapshots, share networks or share groups.
            * description~: The description pattern that can be used to
                filter shares, share snapshots, share networks or share groups.

        :returns: A generator of manila share group resources
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_group.ShareGroup`
        """
        return self._list(_share_group.ShareGroup, **query)

    def get_share_group(self, share_group_id):
        """Lists details for a share group.

        :param share: The ID of the share group to get
        :returns: Details of the identified share group
        :rtype: :class:`~openstack.shared_file_system.v2.
                                    share_group.ShareGroup`
        """
        return self._get(_share_group.ShareGroup, share_group_id)

    def find_share_group(self, name_or_id, ignore_missing=True):
        """Finds a single share group

        :param name_or_id: The name or ID of a share group.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :returns: One :class:`~openstack.shared_file_system.v2.
                                        share_group.ShareGroup`
                                        or None
        """
        return self._find(
            _share_group.ShareGroup, name_or_id, ignore_missing=ignore_missing
        )

    def create_share_group(self, **attrs):
        """Creates a share group from attributes

        :returns: Details of the new share group
        :rtype: :class:`~openstack.shared_file_system.v2.
                                    share_group.ShareGroup`
        """
        return self._create(_share_group.ShareGroup, **attrs)

    def update_share_group(self, share_group_id, **kwargs):
        """Updates details of a single share group

        :param share: The ID of the share group
        :returns: Updated details of the identified share group
        :rtype: :class:`~openstack.shared_file_system.v2.
                                    share_group.ShareGroup`
        """
        return self._update(_share_group.ShareGroup, share_group_id, **kwargs)

    def delete_share_group(self, share_group_id, ignore_missing=True):
        """Deletes a single share group

        :param share: The ID of the share group
        :returns: Result of the "delete" on share group
        :rtype: :class:`~openstack.shared_file_system.v2.
                                    share_group.ShareGroup`
        """
        return self._delete(
            _share_group.ShareGroup,
            share_group_id,
            ignore_missing=ignore_missing,
        )

    def storage_pools(self, details=True, **query):
        """Lists all back-end storage pools with details

        :param kwargs query: Optional query parameters to be sent to limit
            the storage pools being returned.  Available parameters include:

            * pool_name: The pool name for the back end.
            * host_name: The host name for the back end.
            * backend_name: The name of the back end.
            * capabilities: The capabilities for the storage back end.
            * share_type: The share type name or UUID.
        :returns: A generator of manila storage pool resources
        :rtype:
            :class:`~openstack.shared_file_system.v2.storage_pool.StoragePool`
        """
        base_path = '/scheduler-stats/pools/detail' if details else None
        return self._list(
            _storage_pool.StoragePool, base_path=base_path, **query
        )

    def user_messages(self, **query):
        """List shared file system user messages

        :param kwargs query: Optional query parameters to be sent to limit
            the messages being returned.  Available parameters include:

            * action_id: The ID of the action during which the message
              was created.
            * detail_id: The ID of the message detail.
            * limit: The maximum number of shares to return.
            * message_level: The message level.
            * offset: The offset to define start point of share or share
              group listing.
            * sort_key: The key to sort a list of messages.
            * sort_dir: The direction to sort a list of shares.
            * project_id: The ID of the project for which the message
              was created.
            * request_id: The ID of the request during which the message
              was created.
            * resource_id: The UUID of the resource for which the message
              was created.
            * resource_type: The type of the resource for which the message
              was created.

        :returns: A generator of user message resources
        :rtype:
            :class:`~openstack.shared_file_system.v2.user_message.UserMessage`
        """
        return self._list(_user_message.UserMessage, **query)

    def get_user_message(self, message_id):
        """List details of a single user message

        :param message_id: The ID of the user message
        :returns: Details of the identified user message
        :rtype:
            :class:`~openstack.shared_file_system.v2.user_message.UserMessage`
        """
        return self._get(_user_message.UserMessage, message_id)

    def delete_user_message(self, message_id, ignore_missing=True):
        """Deletes a single user message

        :param message_id: The ID of the user message
        :returns: Result of the "delete" on the user message
        :rtype:
            :class:`~openstack.shared_file_system.v2.user_message.UserMessage`
        """
        return self._delete(
            _user_message.UserMessage,
            message_id,
            ignore_missing=ignore_missing,
        )

    def limits(self, **query):
        """Lists all share limits.

        :param kwargs query: Optional query parameters to be sent to limit
            the share limits being returned.

        :returns: A generator of manila share limits resources
        :rtype: :class:`~openstack.shared_file_system.v2.limit.Limit`
        """
        return self._list(_limit.Limit, **query)

    def share_snapshots(self, details=True, **query):
        """Lists all share snapshots with details.

        :param kwargs query: Optional query parameters to be sent to limit
            the snapshots being returned.  Available parameters include:

            * project_id: The ID of the user or service making the API request.

        :returns: A generator of manila share snapshot resources
        :rtype:
            :class:`~openstack.shared_file_system.v2.share_snapshot.ShareSnapshot`
        """
        base_path = '/snapshots/detail' if details else None
        return self._list(
            _share_snapshot.ShareSnapshot, base_path=base_path, **query
        )

    def get_share_snapshot(self, snapshot_id):
        """Lists details of a single share snapshot

        :param snapshot_id: The ID of the snapshot to get
        :returns: Details of the identified share snapshot
        :rtype:
            :class:`~openstack.shared_file_system.v2.share_snapshot.ShareSnapshot`
        """
        return self._get(_share_snapshot.ShareSnapshot, snapshot_id)

    def create_share_snapshot(self, **attrs):
        """Creates a share snapshot from attributes

        :returns: Details of the new share snapshot
        :rtype:
            :class:`~openstack.shared_file_system.v2.share_snapshot.ShareSnapshot`
        """
        return self._create(_share_snapshot.ShareSnapshot, **attrs)

    def update_share_snapshot(self, snapshot_id, **attrs):
        """Updates details of a single share.

        :param snapshot_id: The ID of the snapshot to update
        :pram dict attrs: The attributes to update on the snapshot
        :returns: the updated share snapshot
        :rtype:
            :class:`~openstack.shared_file_system.v2.share_snapshot.ShareSnapshot`
        """
        return self._update(
            _share_snapshot.ShareSnapshot, snapshot_id, **attrs
        )

    def delete_share_snapshot(self, snapshot_id, ignore_missing=True):
        """Deletes a single share snapshot

        :param snapshot_id: The ID of the snapshot to delete
        :returns: Result of the ``delete``
        :rtype: ``None``
        """
        self._delete(
            _share_snapshot.ShareSnapshot,
            snapshot_id,
            ignore_missing=ignore_missing,
        )

    # ========= Network Subnets ==========
    def share_network_subnets(self, share_network_id):
        """Lists all share network subnets with details.

        :param share_network_id: The id of the share network for which
            Share Network Subnets should be listed.
        :returns: A generator of manila share network subnets
        :rtype:
            :class:`~openstack.shared_file_system.v2.share_network_subnet.ShareNetworkSubnet`
        """
        return self._list(
            _share_network_subnet.ShareNetworkSubnet,
            share_network_id=share_network_id,
        )

    def get_share_network_subnet(
        self,
        share_network_id,
        share_network_subnet_id,
    ):
        """Lists details of a single share network subnet.

        :param share_network_id: The id of the share network associated
            with the Share Network Subnet.
        :param share_network_subnet_id: The id of the Share Network Subnet
            to retrieve.
        :returns: Details of the identified share network subnet
        :rtype:
            :class:`~openstack.shared_file_system.v2.share_network_subnet.ShareNetworkSubnet`
        """

        return self._get(
            _share_network_subnet.ShareNetworkSubnet,
            share_network_subnet_id,
            share_network_id=share_network_id,
        )

    def create_share_network_subnet(self, share_network_id, **attrs):
        """Creates a share network subnet from attributes

        :param share_network_id: The id of the share network wthin which the
            the Share Network Subnet should be created.
        :param dict attrs: Attributes which will be used to create
            a share network subnet.
        :returns: Details of the new share network subnet.
        :rtype:
            :class:`~openstack.shared_file_system.v2.share_network_subnet.ShareNetworkSubnet`
        """
        return self._create(
            _share_network_subnet.ShareNetworkSubnet,
            **attrs,
            share_network_id=share_network_id,
        )

    def delete_share_network_subnet(
        self, share_network_id, share_network_subnet, ignore_missing=True
    ):
        """Deletes a share network subnet.

        :param share_network_id: The id of the Share Network associated with
            the Share Network Subnet.
        :param share_network_subnet: The id of the Share Network Subnet
            which should be deleted.
        :returns: Result of the ``delete``
        :rtype: None
        """

        self._delete(
            _share_network_subnet.ShareNetworkSubnet,
            share_network_subnet,
            share_network_id=share_network_id,
            ignore_missing=ignore_missing,
        )

    def share_snapshot_instances(self, details=True, **query):
        """Lists all share snapshot instances with details.

        :param bool details: Whether to fetch detailed resource
            descriptions. Defaults to True.
        :param kwargs query: Optional query parameters to be sent to limit
            the share snapshot instance being returned.
            Available parameters include:

            * snapshot_id: The UUID of the share's base snapshot to filter
                the request based on.
            * project_id: The project ID of the user or service making the
                request.

        :returns: A generator of share snapshot instance resources
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_snapshot_instance.ShareSnapshotInstance`
        """
        base_path = '/snapshot-instances/detail' if details else None
        return self._list(
            _share_snapshot_instance.ShareSnapshotInstance,
            base_path=base_path,
            **query,
        )

    def get_share_snapshot_instance(self, snapshot_instance_id):
        """Lists details of a single share snapshot instance

        :param snapshot_instance_id: The ID of the snapshot instance to get
        :returns: Details of the identified snapshot instance
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_snapshot_instance.ShareSnapshotInstance`
        """
        return self._get(
            _share_snapshot_instance.ShareSnapshotInstance,
            snapshot_instance_id,
        )

    def share_networks(self, details=True, **query):
        """Lists all share networks with details.

        :param dict query: Optional query parameters to be sent to limit the
            resources being returned. Available parameters include:

            * name~: The user defined name of the resource to filter resources
              by.
            * project_id: The ID of the user or service making the request.
            * description~: The description pattern that can be used to filter
              shares, share snapshots, share networks or share groups.
            * all_projects: (Admin only). Defines whether to list the requested
              resources for all projects.

        :returns: Details of shares networks
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_network.ShareNetwork`
        """
        base_path = '/share-networks/detail' if details else None
        return self._list(
            _share_network.ShareNetwork, base_path=base_path, **query
        )

    def get_share_network(self, share_network_id):
        """Lists details of a single share network

        :param share_network: The ID of the share network to get
        :returns: Details of the identified share network
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_network.ShareNetwork`
        """
        return self._get(_share_network.ShareNetwork, share_network_id)

    def delete_share_network(self, share_network_id, ignore_missing=True):
        """Deletes a single share network

        :param share_network_id: The ID of the share network to delete
        :rtype: ``None``
        """
        self._delete(
            _share_network.ShareNetwork,
            share_network_id,
            ignore_missing=ignore_missing,
        )

    def update_share_network(self, share_network_id, **attrs):
        """Updates details of a single share network.

        :param share_network_id: The ID of the share network to update
        :pram dict attrs: The attributes to update on the share network
        :returns: the updated share network
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_network.ShareNetwork`
        """
        return self._update(
            _share_network.ShareNetwork, share_network_id, **attrs
        )

    def create_share_network(self, **attrs):
        """Creates a share network from attributes

        :returns: Details of the new share network
        :param dict attrs: Attributes which will be used to create
            a :class:`~openstack.shared_file_system.v2.
            share_network.ShareNetwork`,comprised of the properties
            on the ShareNetwork class.
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_network.ShareNetwork`
        """
        return self._create(_share_network.ShareNetwork, **attrs)

    def share_instances(self, **query):
        """Lists all share instances.

        :param kwargs query: Optional query parameters to be sent to limit
            the share instances being returned. Available parameters include:

        * export_location_id: The export location UUID that can be used
          to filter share instances.
        * export_location_path: The export location path that can be used
          to filter share instances.

        :returns: Details of share instances resources
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_instance.ShareInstance`
        """
        return self._list(_share_instance.ShareInstance, **query)

    def get_share_instance(self, share_instance_id):
        """Shows details for a single share instance

        :param share_instance_id: The UUID of the share instance to get

        :returns: Details of the identified share instance
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_instance.ShareInstance`
        """
        return self._get(_share_instance.ShareInstance, share_instance_id)

    def reset_share_instance_status(self, share_instance_id, status):
        """Explicitly updates the state of a share instance.

        :param share_instance_id: The UUID of the share instance to reset.
        :param status: The share or share instance status to be set.

        :returns: ``None``
        """
        res = self._get_resource(
            _share_instance.ShareInstance, share_instance_id
        )
        res.reset_status(self, status)

    def delete_share_instance(self, share_instance_id):
        """Force-deletes a share instance

        :param share_instance: The ID of the share instance to delete

        :returns: ``None``
        """
        res = self._get_resource(
            _share_instance.ShareInstance, share_instance_id
        )
        res.force_delete(self)

    def export_locations(self, share_id):
        """List all export locations with details

        :param share_id: The ID of the share to list export locations from
        :returns: List of export locations
        :rtype: List of :class:`~openstack.shared_filesystem_storage.v2.
            share_export_locations.ShareExportLocations`
        """
        return self._list(
            _share_export_locations.ShareExportLocation, share_id=share_id
        )

    def get_export_location(self, export_location, share_id):
        """List details of export location

        :param export_location: The export location resource to get
        :param share_id: The ID of the share to get export locations from
        :returns: Details of identified export location
        :rtype: :class:`~openstack.shared_filesystem_storage.v2.
            share_export_locations.ShareExportLocations`
        """

        export_location_id = resource.Resource._get_id(export_location)
        return self._get(
            _share_export_locations.ShareExportLocation,
            export_location_id,
            share_id=share_id,
        )

    def access_rules(self, share, **query):
        """Lists the access rules on a share.

        :returns: A generator of the share access rules.
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_access_rules.ShareAccessRules`
        """
        share = self._get_resource(_share.Share, share)
        return self._list(
            _share_access_rule.ShareAccessRule, share_id=share.id, **query
        )

    def get_access_rule(self, access_id):
        """List details of an access rule.

        :param access_id: The id of the access rule to get
        :returns: Details of the identified access rule.
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_access_rules.ShareAccessRules`
        """
        return self._get(_share_access_rule.ShareAccessRule, access_id)

    def create_access_rule(self, share_id, **attrs):
        """Creates an access rule from attributes

        :returns: Details of the new access rule
        :param share_id: The ID of the share
        :param dict attrs: Attributes which will be used to create
            a :class:`~openstack.shared_file_system.v2.
            share_access_rules.ShareAccessRules`, comprised of the
            properties on the ShareAccessRules class.
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_access_rules.ShareAccessRules`
        """
        base_path = f"/shares/{share_id}/action"
        return self._create(
            _share_access_rule.ShareAccessRule, base_path=base_path, **attrs
        )

    def delete_access_rule(
        self, access_id, share_id, ignore_missing=True, *, unrestrict=False
    ):
        """Deletes an access rule

        :param access_id: The id of the access rule to get
        :param share_id: The ID of the share
        :param unrestrict: If Manila must attempt removing locks while deleting

        :rtype: ``requests.models.Response`` HTTP response from internal
            requests client
        """
        res = self._get_resource(
            _share_access_rule.ShareAccessRule, access_id, share_id=share_id
        )
        try:
            return res.delete(
                self,
                unrestrict=unrestrict,
            )
        except exceptions.NotFoundException:
            if ignore_missing:
                return None
            raise

    def share_group_snapshots(self, details=True, **query):
        """Lists all share group snapshots.

        :param kwargs query: Optional query parameters to be sent
            to limit the share group snapshots being returned.
            Available parameters include:

            * project_id: The ID of the project that owns the resource.
            * name: The user defined name of the resource to filter resources.
            * description: The user defined description text that can be used
              to filter resources.
            * status: Filters by a share status
            * share_group_id: The UUID of a share group to filter resource.
            * limit: The maximum number of share group snapshot members
              to return.
            * offset: The offset to define start point of share or
              share group listing.
            * sort_key: The key to sort a list of shares.
            * sort_dir: The direction to sort a list of shares. A valid
              value is asc, or desc.

        :returns: Details of share group snapshots resources
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_group_snapshot.ShareGroupSnapshot`
        """
        base_path = '/share-group-snapshots/detail' if details else None
        return self._list(
            _share_group_snapshot.ShareGroupSnapshot,
            base_path=base_path,
            **query,
        )

    def share_group_snapshot_members(self, group_snapshot_id):
        """Lists all share group snapshots members.

        :param group_snapshot_id: The ID of the group snapshot to get
        :returns: List of the share group snapshot members, which are
            share snapshots.
        :rtype: dict containing attributes of the share snapshot members.
        """
        res = self._get(
            _share_group_snapshot.ShareGroupSnapshot,
            group_snapshot_id,
        )
        response = res.get_members(self)
        return response

    def get_share_group_snapshot(self, group_snapshot_id):
        """Show share group snapshot details

        :param group_snapshot_id: The ID of the group snapshot to get
        :returns: Details of the group snapshot
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_group_snapshot.ShareGroupSnapshot`
        """
        return self._get(
            _share_group_snapshot.ShareGroupSnapshot, group_snapshot_id
        )

    def create_share_group_snapshot(self, share_group_id, **attrs):
        """Creates a point-in-time snapshot copy of a share group.

        :returns: Details of the new snapshot
        :param dict attrs: Attributes which will be used to create
            a :class:`~openstack.shared_file_system.v2.
            share_group_snapshots.ShareGroupSnapshots`,
        :param 'share_group_id': ID of the share group to have the snapshot
            taken.
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_group_snapshot.ShareGroupSnapshot`
        """
        return self._create(
            _share_group_snapshot.ShareGroupSnapshot,
            share_group_id=share_group_id,
            **attrs,
        )

    def reset_share_group_snapshot_status(self, group_snapshot_id, status):
        """Reset share group snapshot state.

        :param group_snapshot_id: The ID of the share group snapshot to reset
        :param status: The state of the share group snapshot to be set, A
            valid value is "creating", "error", "available", "deleting",
            "error_deleting".
        :rtype: ``None``
        """
        res = self._get(
            _share_group_snapshot.ShareGroupSnapshot, group_snapshot_id
        )
        res.reset_status(self, status)

    def update_share_group_snapshot(self, group_snapshot_id, **attrs):
        """Updates a share group snapshot.

        :param group_snapshot_id: The ID of the share group snapshot to update
        :param dict attrs: The attributes to update on the share group snapshot
        :returns: the updated share group snapshot
        :rtype: :class:`~openstack.shared_file_system.v2.
            share_group_snapshot.ShareGroupSnapshot`
        """
        return self._update(
            _share_group_snapshot.ShareGroupSnapshot,
            group_snapshot_id,
            **attrs,
        )

    def delete_share_group_snapshot(
        self, group_snapshot_id, ignore_missing=True
    ):
        """Deletes a share group snapshot.

        :param group_snapshot_id: The ID of the share group snapshot to delete
        :rtype: ``None``
        """
        self._delete(
            _share_group_snapshot.ShareGroupSnapshot,
            group_snapshot_id,
            ignore_missing=ignore_missing,
        )

    # ========= Share Metadata ==========
    def get_share_metadata(self, share_id):
        """Lists all metadata for a share.

        :param share_id: The ID of the share

        :returns: A :class:`~openstack.shared_file_system.v2.share.Share`
            with the share's metadata.
        :rtype:
            :class:`~openstack.shared_file_system.v2.share.Share`
        """
        share = self._get_resource(_share.Share, share_id)
        return share.fetch_metadata(self)

    def get_share_metadata_item(self, share_id, key):
        """Retrieves a specific metadata item from a share by its key.

        :param share_id: The ID of the share
        :param key: The key of the share metadata

        :returns: A :class:`~openstack.shared_file_system.v2.share.Share`
            with the share's metadata.
        :rtype:
            :class:`~openstack.shared_file_system.v2.share.Share`
        """
        share = self._get_resource(_share.Share, share_id)
        return share.get_metadata_item(self, key)

    def create_share_metadata(self, share_id, **metadata):
        """Creates share metadata as key-value pairs.

        :param share_id: The ID of the share
        :param metadata: The metadata to be created

        :returns: A :class:`~openstack.shared_file_system.v2.share.Share`
            with the share's metadata.
        :rtype:
            :class:`~openstack.shared_file_system.v2.share.Share`
        """
        share = self._get_resource(_share.Share, share_id)
        return share.set_metadata(self, metadata=metadata)

    def update_share_metadata(self, share_id, metadata, replace=False):
        """Updates metadata of given share.

        :param share_id: The ID of the share
        :param metadata: The metadata to be created
        :param replace: Boolean for whether the preexisting metadata
            should be replaced

        :returns: A :class:`~openstack.shared_file_system.v2.share.Share`
            with the share's updated metadata.
        :rtype:
            :class:`~openstack.shared_file_system.v2.share.Share`
        """
        share = self._get_resource(_share.Share, share_id)
        return share.set_metadata(self, metadata=metadata, replace=replace)

    def delete_share_metadata(self, share_id, keys, ignore_missing=True):
        """Deletes a single metadata item on a share, idetified by its key.

        :param share_id: The ID of the share
        :param keys: The list of share metadata keys to be deleted
        :param ignore_missing: Boolean indicating if missing keys should be
            ignored.

        :returns: None
        :rtype: None
        """
        share = self._get_resource(_share.Share, share_id)
        keys_failed_to_delete = []
        for key in keys:
            try:
                share.delete_metadata_item(self, key)
            except exceptions.NotFoundException:
                if not ignore_missing:
                    self._connection.log.info("Key %s not found.", key)
                    keys_failed_to_delete.append(key)
            except exceptions.ForbiddenException:
                self._connection.log.info("Key %s cannot be deleted.", key)
                keys_failed_to_delete.append(key)
            except exceptions.SDKException:
                self._connection.log.info("Failed to delete key %s.", key)
                keys_failed_to_delete.append(key)
        if keys_failed_to_delete:
            raise exceptions.SDKException(
                f"Some keys failed to be deleted {keys_failed_to_delete}"
            )

    def resource_locks(self, **query):
        """Lists all resource locks.

        :param kwargs query: Optional query parameters to be sent to limit
            the resource locks being returned.  Available parameters include:

            * project_id: The project ID of the user that the lock is
                created for.
            * user_id: The ID of a user to filter resource locks by.
            * all_projects: list locks from all projects (Admin Only)
            * resource_id: The ID of the resource that the locks pertain to
                filter resource locks by.
            * resource_action: The action prevented by the filtered resource
                locks.
            * resource_type: The type of the resource that the locks pertain
                to filter resource locks by.
            * lock_context: The lock creator's context to filter locks by.
            * lock_reason: The lock reason that can be used to filter resource
                locks. (Inexact search is also available with lock_reason~)
            * created_since: Search for the list of resources that were created
                after the specified date. The date is in 'yyyy-mm-dd' format.
            * created_before: Search for the list of resources that were
                created prior to the specified date. The date is in
                'yyyy-mm-dd' format.
            * limit: The maximum number of resource locks to return.
            * offset: The offset to define start point of resource lock
                listing.
            * sort_key: The key to sort a list of shares.
            * sort_dir: The direction to sort a list of shares
            * with_count: Whether to show count in API response or not,
                default is False. This query parameter is useful with
                pagination.

        :returns: A generator of manila resource locks
        :rtype: :class:`~openstack.shared_file_system.v2.
            resource_locks.ResourceLock`
        """

        if query.get('resource_type'):
            # The _create method has a parameter named resource_type, which
            # refers to the type of resource to be created, so we need to avoid
            # a conflict of parameters we are sending to the method.
            query['__conflicting_attrs'] = {
                'resource_type': query.get('resource_type')
            }
            query.pop('resource_type')
        return self._list(_resource_locks.ResourceLock, **query)

    def get_resource_lock(self, resource_lock):
        """Show details of a resource lock.

        :param resource_lock: The ID of a resource lock or a
            :class:`~openstack.shared_file_system.v2.
            resource_locks.ResourceLock` instance.
        :returns: Details of the identified resource lock.
        :rtype: :class:`~openstack.shared_file_system.v2.
            resource_locks.ResourceLock`
        """
        return self._get(_resource_locks.ResourceLock, resource_lock)

    def update_resource_lock(self, resource_lock, **attrs):
        """Updates details of a single resource lock.

        :param resource_lock: The ID of a resource lock or a
            :class:`~openstack.shared_file_system.v2.
            resource_locks.ResourceLock` instance.
        :param dict attrs: The attributes to update on the resource lock
        :returns: the updated resource lock
        :rtype: :class:`~openstack.shared_file_system.v2.
            resource_locks.ResourceLock`
        """
        return self._update(
            _resource_locks.ResourceLock, resource_lock, **attrs
        )

    def delete_resource_lock(self, resource_lock, ignore_missing=True):
        """Deletes a single resource lock

        :param resource_lock: The ID of a resource lock or a
            :class:`~openstack.shared_file_system.v2.
            resource_locks.ResourceLock` instance.
        :returns: Result of the ``delete``
        :rtype: ``None``
        """
        return self._delete(
            _resource_locks.ResourceLock,
            resource_lock,
            ignore_missing=ignore_missing,
        )

    def create_resource_lock(self, **attrs):
        """Locks a resource.

        :param dict attrs: Attributes which will be used to create
            a :class:`~openstack.shared_file_system.v2.
            resource_locks.ResourceLock`, comprised of the properties
            on the ResourceLock class. Available parameters include:

            * ``resource_id``: ID of the resource to be locked.
            * ``resource_type``: type of the resource (share, access_rule).
            * ``resource_action``: action to be locked (delete, show).
            * ``lock_reason``: reason why you're locking the resource
                (Optional).
        :returns: Details of the lock
        :rtype: :class:`~openstack.shared_file_system.v2.
            resource_locks.ResourceLock`
        """

        if attrs.get('resource_type'):
            # The _create method has a parameter named resource_type, which
            # refers to the type of resource to be created, so we need to avoid
            # a conflict of parameters we are sending to the method.
            attrs['__conflicting_attrs'] = {
                'resource_type': attrs.get('resource_type')
            }
            attrs.pop('resource_type')
        return self._create(_resource_locks.ResourceLock, **attrs)

    def get_quota_class_set(self, quota_class_name):
        """Get quota class set.

        :param quota_class_name: The name of the quota class
        :returns: A :class:`~openstack.shared_file_system.v2
            .quota_class_set.QuotaClassSet`
        """
        return self._get(_quota_class_set.QuotaClassSet, quota_class_name)

    def update_quota_class_set(self, quota_class_name, **attrs):
        """Update quota class set.

        :param quota_class_name: The name of the quota class
        :param attrs: The attributes to update on the quota class set
        :returns: the updated quota class set
        :rtype: :class:`~openstack.shared_file_system.v2
            .quota_class_set.QuotaClassSet`
        """

        return self._update(
            _quota_class_set.QuotaClassSet, quota_class_name, **attrs
        )

    # ========== Utilities ==========

    def wait_for_status(
        self,
        res: resource.ResourceT,
        status: str,
        failures: list[str] | None = None,
        interval: int | float | None = 2,
        wait: int | None = None,
        attribute: str = 'status',
        callback: ty.Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for the resource to be in a particular status.

        :param session: The session to use for making this request.
        :param resource: The resource to wait on to reach the status. The
            resource must have a status attribute specified via ``attribute``.
        :param status: Desired status of the resource.
        :param failures: Statuses that would indicate the transition
            failed such as 'ERROR'. Defaults to ['ERROR'].
        :param interval: Number of seconds to wait between checks.
        :param wait: Maximum number of seconds to wait for transition.
            Set to ``None`` to wait forever.
        :param attribute: Name of the resource attribute that contains the
            status.
        :param callback: A callback function. This will be called with a single
            value, progress. This is API specific but is generally a percentage
            value from 0-100.

        :return: The updated resource.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if the
            transition to status failed to occur in ``wait`` seconds.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
            transitioned to one of the states in ``failures``.
        :raises: :class:`~AttributeError` if the resource does not have a
            ``status`` attribute
        """
        return resource.wait_for_status(
            self, res, status, failures, interval, wait, attribute, callback
        )

    def wait_for_delete(
        self,
        res: resource.ResourceT,
        interval: int = 2,
        wait: int = 120,
        callback: ty.Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for a resource to be deleted.

        :param res: The resource to wait on to be deleted.
        :param interval: Number of seconds to wait before to consecutive
            checks.
        :param wait: Maximum number of seconds to wait before the change.
        :param callback: A callback function. This will be called with a single
            value, progress, which is a percentage value from 0-100.

        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
            to delete failed to occur in the specified seconds.
        """
        return resource.wait_for_delete(self, res, interval, wait, callback)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from openstack.container_infrastructure_management.v1 import (
    cluster as _cluster,
)
from openstack.container_infrastructure_management.v1 import (
    cluster_certificate as _cluster_cert,
)
from openstack.container_infrastructure_management.v1 import (
    cluster_template as _cluster_template,
)
from openstack.container_infrastructure_management.v1 import (
    service as _service,
)
from openstack import proxy
from openstack import resource


class Proxy(proxy.Proxy):
    _resource_registry = {
        "cluster": _cluster.Cluster,
        "cluster_template": _cluster_template.ClusterTemplate,
        "service": _service.Service,
    }

    # ========== Clusters ==========

    def create_cluster(self, **attrs):
        """Create a new cluster from attributes

        :param dict attrs: Keyword arguments which will be used to create a
            :class:`~openstack.container_infrastructure_management.v1.cluster.Cluster`,
            comprised of the properties on the Cluster class.
        :returns: The results of cluster creation
        :rtype:
            :class:`~openstack.container_infrastructure_management.v1.cluster.Cluster`
        """
        return self._create(_cluster.Cluster, **attrs)

    def delete_cluster(self, cluster, ignore_missing=True):
        """Delete a cluster

        :param cluster: The value can be either the ID of a cluster or a
            :class:`~openstack.container_infrastructure_management.v1.cluster.Cluster`
            instance.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the cluster does not exist. When set to ``True``, no exception
            will be set when attempting to delete a nonexistent cluster.
        :returns: ``None``
        """
        self._delete(_cluster.Cluster, cluster, ignore_missing=ignore_missing)

    def find_cluster(self, name_or_id, ignore_missing=True):
        """Find a single cluster

        :param name_or_id: The name or ID of a cluster.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :returns: One
            :class:`~openstack.container_infrastructure_management.v1.cluster.Cluster`
            or None
        """
        return self._find(
            _cluster.Cluster,
            name_or_id,
            ignore_missing=ignore_missing,
        )

    def get_cluster(self, cluster):
        """Get a single cluster

        :param cluster: The value can be the ID of a cluster or a
            :class:`~openstack.container_infrastructure_management.v1.cluster.Cluster`
            instance.

        :returns: One
            :class:`~openstack.container_infrastructure_management.v1.cluster.Cluster`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(_cluster.Cluster, cluster)

    def clusters(self, **query):
        """Return a generator of clusters

        :param kwargs query: Optional query parameters to be sent to limit
            the resources being returned.

        :returns: A generator of cluster objects
        :rtype:
            :class:`~openstack.container_infrastructure_management.v1.cluster.Cluster`
        """
        return self._list(_cluster.Cluster, **query)

    def update_cluster(self, cluster, **attrs):
        """Update a cluster

        :param cluster: Either the id of a cluster or a
            :class:`~openstack.container_infrastructure_management.v1.cluster.Cluster`
            instance.
        :param attrs: The attributes to update on the cluster represented
            by ``cluster``.

        :returns: The updated cluster
        :rtype:
            :class:`~openstack.container_infrastructure_management.v1.cluster.Cluster`
        """
        return self._update(_cluster.Cluster, cluster, **attrs)

    # ============== Cluster Templates ==============

    def create_cluster_template(self, **attrs):
        """Create a new cluster_template from attributes

        :param dict attrs: Keyword arguments which will be used to create a
            :class:`~openstack.container_infrastructure_management.v1.cluster_template.ClusterTemplate`,
            comprised of the properties on the ClusterTemplate class.
        :returns: The results of cluster_template creation
        :rtype:
            :class:`~openstack.container_infrastructure_management.v1.cluster_template.ClusterTemplate`
        """
        return self._create(_cluster_template.ClusterTemplate, **attrs)

    def delete_cluster_template(self, cluster_template, ignore_missing=True):
        """Delete a cluster_template

        :param cluster_template: The value can be either the ID of a
            cluster_template or a
            :class:`~openstack.container_infrastructure_management.v1.cluster_template.ClusterTemplate`
            instance.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the cluster_template does not exist. When set to ``True``, no
            exception will be set when attempting to delete a nonexistent
            cluster_template.
        :returns: ``None``
        """
        self._delete(
            _cluster_template.ClusterTemplate,
            cluster_template,
            ignore_missing=ignore_missing,
        )

    def find_cluster_template(self, name_or_id, ignore_missing=True):
        """Find a single cluster_template

        :param name_or_id: The name or ID of a cluster_template.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :returns: One
            :class:`~openstack.container_infrastructure_management.v1.cluster_template.ClusterTemplate`
            or None
        """
        return self._find(
            _cluster_template.ClusterTemplate,
            name_or_id,
            ignore_missing=ignore_missing,
        )

    def get_cluster_template(self, cluster_template):
        """Get a single cluster_template

        :param cluster_template: The value can be the ID of a cluster_template
            or a
            :class:`~openstack.container_infrastructure_management.v1.cluster_template.ClusterTemplate`
            instance.

        :returns: One
            :class:`~openstack.container_infrastructure_management.v1.cluster_template.ClusterTemplate`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(_cluster_template.ClusterTemplate, cluster_template)

    def cluster_templates(self, **query):
        """Return a generator of cluster_templates

        :param kwargs query: Optional query parameters to be sent to limit
            the resources being returned.

        :returns: A generator of cluster_template objects
        :rtype:
            :class:`~openstack.container_infrastructure_management.v1.cluster_template.ClusterTemplate`
        """
        return self._list(_cluster_template.ClusterTemplate, **query)

    def update_cluster_template(self, cluster_template, **attrs):
        """Update a cluster_template

        :param cluster_template: Either the id of a cluster_template or a
            :class:`~openstack.container_infrastructure_management.v1.cluster_template.ClusterTemplate`
            instance.
        :param attrs: The attributes to update on the cluster_template
            represented by ``cluster_template``.

        :returns: The updated cluster_template
        :rtype:
            :class:`~openstack.container_infrastructure_management.v1.cluster_template.ClusterTemplate`
        """
        return self._update(
            _cluster_template.ClusterTemplate, cluster_template, **attrs
        )

    # ============== Cluster Certificates ==============

    def create_cluster_certificate(self, **attrs):
        """Create a new cluster_certificate from CSR

        :param dict attrs: Keyword arguments which will be used to create a
            :class:`~openstack.container_infrastructure_management.v1.cluster_certificate.ClusterCertificate`,
            comprised of the properties on the ClusterCertificate class.
        :returns: The results of cluster_certificate creation
        :rtype:
            :class:`~openstack.container_infrastructure_management.v1.cluster_certificate.ClusterCertificate`
        """
        return self._create(_cluster_cert.ClusterCertificate, **attrs)

    def get_cluster_certificate(self, cluster_certificate):
        """Get a single cluster_certificate

        :param cluster_certificate: The value can be the ID of a
            cluster_certificate or a
            :class:`~openstack.container_infrastructure_management.v1.cluster_certificate.ClusterCertificate`
            instance.

        :returns: One
            :class:`~openstack.container_infrastructure_management.v1.cluster_certificate.ClusterCertificate`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._get(_cluster_cert.ClusterCertificate, cluster_certificate)

    # ============== Services ==============

    def services(self):
        """Return a generator of services

        :returns: A generator of service objects
        :rtype:
            :class:`~openstack.container_infrastructure_management.v1.service.Service`
        """
        return self._list(_service.Service)

    # ========== Utilities ==========

    def wait_for_status(
        self,
        res: resource.ResourceT,
        status: str,
        failures: list[str] | None = None,
        interval: int | float | None = 2,
        wait: int | None = None,
        attribute: str = 'status',
        callback: ty.Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for the resource to be in a particular status.

        :param session: The session to use for making this request.
        :param resource: The resource to wait on to reach the status. The
            resource must have a status attribute specified via ``attribute``.
        :param status: Desired status of the resource.
        :param failures: Statuses that would indicate the transition
            failed such as 'ERROR'. Defaults to ['ERROR'].
        :param interval: Number of seconds to wait between checks.
        :param wait: Maximum number of seconds to wait for transition.
            Set to ``None`` to wait forever.
        :param attribute: Name of the resource attribute that contains the
            status.
        :param callback: A callback function. This will be called with a single
            value, progress. This is API specific but is generally a percentage
            value from 0-100.

        :return: The updated resource.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if the
            transition to status failed to occur in ``wait`` seconds.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
            transitioned to one of the states in ``failures``.
        :raises: :class:`~AttributeError` if the resource does not have a
            ``status`` attribute
        """
        return resource.wait_for_status(
            self, res, status, failures, interval, wait, attribute, callback
        )

    def wait_for_delete(
        self,
        res: resource.ResourceT,
        interval: int = 2,
        wait: int = 120,
        callback: ty.Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for a resource to be deleted.

        :param res: The resource to wait on to be deleted.
        :param interval: Number of seconds to wait before to consecutive
            checks.
        :param wait: Maximum number of seconds to wait before the change.
        :param callback: A callback function. This will be called with a single
            value, progress, which is a percentage value from 0-100.

        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
            to delete failed to occur in the specified seconds.
        """
        return resource.wait_for_delete(self, res, interval, wait, callback)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
fixtures
--------

Fixtures used for testing
"""

import warnings

import fixtures

from openstack import warnings as os_warnings


# TODO(stephenfin): Replace this with WarningsFilter from fixtures when it's
# released https://github.com/testing-cabal/fixtures/pull/50
class WarningsFixture(fixtures.Fixture):
    """Filters out warnings during test runs."""

    def setUp(self):
        super().setUp()

        self._original_warning_filters = warnings.filters[:]

        # enable user warnings as many libraries use this (it's the default)
        warnings.simplefilter("error", UserWarning)

        # enable deprecation warnings in general...
        warnings.simplefilter("once", DeprecationWarning)

        # ...but ignore our own deprecation warnings
        warnings.filterwarnings(
            "ignore",
            category=os_warnings.OpenStackDeprecationWarning,
        )
        warnings.filterwarnings(
            "ignore",
            category=os_warnings._RemovedInSDKWarning,
        )

        # also ignore our own general warnings
        warnings.filterwarnings(
            "ignore",
            category=os_warnings.OpenStackWarning,
        )

        self.addCleanup(self._reset_warning_filters)

    def _reset_warning_filters(self):
        warnings.filters[:] = self._original_warning_filters  # type: ignore[index]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_update_server
----------------------------------

Tests for the `update_server` command.
"""

import uuid

from openstack import exceptions
from openstack.tests import fakes
from openstack.tests.unit import base


class TestUpdateServer(base.TestCase):
    def setUp(self):
        super().setUp()
        self.server_id = str(uuid.uuid4())
        self.server_name = self.getUniqueString('name')
        self.updated_server_name = self.getUniqueString('name2')
        self.fake_server = fakes.make_fake_server(
            self.server_id, self.server_name
        )

    def test_update_server_with_update_exception(self):
        """
        Test that an exception in the update raises an exception in
        update_server.
        """
        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', self.server_name],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', 'detail'],
                        qs_elements=[f'name={self.server_name}'],
                    ),
                    json={'servers': [self.fake_server]},
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['servers', self.server_id]
                    ),
                    status_code=400,
                    validate=dict(
                        json={'server': {'name': self.updated_server_name}}
                    ),
                ),
            ]
        )
        self.assertRaises(
            exceptions.SDKException,
            self.cloud.update_server,
            self.server_name,
            name=self.updated_server_name,
        )

        self.assert_calls()

    def test_update_server_name(self):
        """
        Test that update_server updates the name without raising any exception
        """
        fake_update_server = fakes.make_fake_server(
            self.server_id, self.updated_server_name
        )

        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', self.server_name],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['servers', 'detail'],
                        qs_elements=[f'name={self.server_name}'],
                    ),
                    json={'servers': [self.fake_server]},
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['servers', self.server_id]
                    ),
                    json={'server': fake_update_server},
                    validate=dict(
                        json={'server': {'name': self.updated_server_name}}
                    ),
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'networks': []},
                ),
            ]
        )
        self.assertEqual(
            self.updated_server_name,
            self.cloud.update_server(
                self.server_name, name=self.updated_server_name
            )['name'],
        )

        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack import exceptions
from openstack.image.v2 import service_info as si
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE_IMPORT = {
    'import-methods': {
        'description': 'Import methods available.',
        'type': 'array',
        'value': ['glance-direct', 'web-download'],
    }
}
EXAMPLE_STORE = {
    'id': IDENTIFIER,
    'description': 'Fast access to rbd store',
    'default': True,
    'properties': {
        "pool": "pool1",
        "chunk_size": 65536,
        "thin_provisioning": False,
    },
}


class TestStore(base.TestCase):
    def test_basic(self):
        sot = si.Store()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('stores', sot.resources_key)
        self.assertEqual('/info/stores', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertFalse(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = si.Store(**EXAMPLE_STORE)
        self.assertEqual(EXAMPLE_STORE['id'], sot.id)
        self.assertEqual(EXAMPLE_STORE['description'], sot.description)
        self.assertEqual(EXAMPLE_STORE['default'], sot.is_default)
        self.assertEqual(EXAMPLE_STORE['properties'], sot.properties)

    @mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
    def test_delete_image(self):
        sot = si.Store(**EXAMPLE_STORE)
        session = mock.Mock()
        session.delete = mock.Mock()

        sot.delete_image(session, image='image_id')

        session.delete.assert_called_with('stores/IDENTIFIER/image_id')


class TestImport(base.TestCase):
    def test_basic(self):
        sot = si.Import()
        self.assertIsNone(sot.resource_key)
        self.assertIsNone(sot.resources_key)
        self.assertEqual('/info/import', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertFalse(sot.allow_list)

    def test_make_it(self):
        sot = si.Import(**EXAMPLE_IMPORT)
        self.assertEqual(EXAMPLE_IMPORT['import-methods'], sot.import_methods)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from examples.image import create as image_create
from examples.image import delete as image_delete
from examples.image import list as image_list

from openstack.tests.functional import base


class TestImage(base.BaseFunctionalTest):
    """Test the image examples

    The purpose of these tests is to ensure the examples run successfully.
    """

    def test_image(self):
        image_list.list_images(self.operator_cloud)

        image_create.upload_image(self.operator_cloud)

        image_delete.delete_image(self.operator_cloud)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.identity.v3 import group as _group
from openstack.identity.v3 import user as _user
from openstack.tests.functional import base


class TestGroup(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()

        self.group_name = self.getUniqueString('group')
        self.group_description = self.getUniqueString('group')
        self.user_name = self.getUniqueString('user')
        self.user_email = f"{self.user_name}@example.com"

        self.user = self.operator_cloud.identity.create_user(
            name=self.user_name,
            email=self.user_email,
        )
        self.addCleanup(self._delete_user, self.user)

    def _delete_group(self, group):
        ret = self.operator_cloud.identity.delete_group(group)
        self.assertIsNone(ret)

    def _delete_user(self, user):
        ret = self.operator_cloud.identity.delete_user(user)
        self.assertIsNone(ret)

    def test_group(self):
        # create the group

        group = self.operator_cloud.identity.create_group(
            name=self.group_name,
        )
        self.addCleanup(self._delete_group, group)
        self.assertIsInstance(group, _group.Group)
        self.assertEqual('', group.description)

        # update the group

        group = self.operator_cloud.identity.update_group(
            group, description=self.group_description
        )
        self.assertIsInstance(group, _group.Group)
        self.assertEqual(self.group_description, group.description)

        # retrieve details of the (updated) group by ID

        group = self.operator_cloud.identity.get_group(group.id)
        self.assertIsInstance(group, _group.Group)
        self.assertEqual(self.group_description, group.description)

        # retrieve details of the (updated) group by name

        group = self.operator_cloud.identity.find_group(group.name)
        self.assertIsInstance(group, _group.Group)
        self.assertEqual(self.group_description, group.description)

        # list all groups

        groups = list(self.operator_cloud.identity.groups())
        self.assertIsInstance(groups[0], _group.Group)
        self.assertIn(
            self.group_name,
            {x.name for x in groups},
        )

        # add user to group
        self.operator_cloud.identity.add_user_to_group(self.user, group)

        is_in_group = self.operator_cloud.identity.check_user_in_group(
            self.user, group
        )
        self.assertTrue(is_in_group)

        group_users = list(self.operator_cloud.identity.group_users(group))
        self.assertIsInstance(group_users[0], _user.User)
        self.assertIn(self.user_name, {x.name for x in group_users})

        # remove user from group

        self.operator_cloud.identity.remove_user_from_group(self.user, group)

        is_in_group = self.operator_cloud.identity.check_user_in_group(
            self.user, group
        )
        self.assertFalse(is_in_group)



#   Copyright 2021 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from openstack.network.v2 import local_ip_association as _local_ip_association
from openstack.tests.functional import base


class TestLocalIPAssociation(base.BaseFunctionalTest):
    LOCAL_IP_ID = None
    FIXED_PORT_ID = None
    FIXED_IP = None

    def setUp(self):
        super().setUp()

        if not self.user_cloud.network.find_extension("local_ip"):
            self.skipTest("Local IP extension disabled")

        self.LOCAL_IP_ID = self.getUniqueString()
        self.FIXED_PORT_ID = self.getUniqueString()
        self.FIXED_IP = self.getUniqueString()
        local_ip_association = (
            self.user_cloud.network.create_local_ip_association(
                local_ip=self.LOCAL_IP_ID,
                fixed_port_id=self.FIXED_PORT_ID,
                fixed_ip=self.FIXED_IP,
            )
        )
        assert isinstance(
            local_ip_association, _local_ip_association.LocalIPAssociation
        )
        self.assertEqual(self.LOCAL_IP_ID, local_ip_association.local_ip_id)
        self.assertEqual(
            self.FIXED_PORT_ID, local_ip_association.fixed_port_id
        )
        self.assertEqual(self.FIXED_IP, local_ip_association.fixed_ip)

    def tearDown(self):
        sot = self.user_cloud.network.delete_local_ip_association(
            self.LOCAL_IP_ID, self.FIXED_PORT_ID
        )
        self.assertIsNone(sot)
        super().tearDown()

    def test_find(self):
        sot = self.user_cloud.network.find_local_ip_association(
            self.FIXED_PORT_ID, self.LOCAL_IP_ID
        )
        self.assertEqual(self.FIXED_PORT_ID, sot.fixed_port_id)

    def test_get(self):
        sot = self.user_cloud.network.get_local_ip_association(
            self.FIXED_PORT_ID, self.LOCAL_IP_ID
        )
        self.assertEqual(self.FIXED_PORT_ID, sot.fixed_port_id)

    def test_list(self):
        fixed_port_id = [
            obj.fixed_port_id
            for obj in self.user_cloud.network.local_ip_associations(
                self.LOCAL_IP_ID
            )
        ]
        self.assertIn(self.FIXED_PORT_ID, fixed_port_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack import exceptions
from openstack.tests.functional import base


# NOTE: method to make mypy happy.
def _get_command(*args):
    return mock.Mock()


class TestTagNeutron(base.BaseFunctionalTest):
    get_command = _get_command

    def test_set_tags(self):
        sot = self.get_command(self.ID)
        self.assertEqual([], sot.tags)

        self.user_cloud.network.set_tags(sot, ["blue"])
        sot = self.get_command(self.ID)
        self.assertEqual(["blue"], sot.tags)

        self.user_cloud.network.set_tags(sot, [])
        sot = self.get_command(self.ID)
        self.assertEqual([], sot.tags)

    def test_get_tags(self):
        sot = self.get_command(self.ID)
        self.assertEqual([], sot.tags)

        self.user_cloud.network.set_tags(sot, ["blue", "red"])
        tags = self.user_cloud.network.get_tags(sot)
        self.assertEqual(["blue", "red"], tags)

    def test_add_tag(self):
        sot = self.get_command(self.ID)
        self.assertEqual([], sot.tags)

        self.user_cloud.network.add_tag(sot, "blue")
        tags = self.user_cloud.network.get_tags(sot)
        self.assertEqual(["blue"], tags)

        # The operation is idempotent.
        self.user_cloud.network.add_tag(sot, "blue")
        tags = self.user_cloud.network.get_tags(sot)
        self.assertEqual(["blue"], tags)

    def test_remove_tag(self):
        sot = self.get_command(self.ID)
        self.assertEqual([], sot.tags)

        self.user_cloud.network.set_tags(sot, ["blue"])
        tags = self.user_cloud.network.get_tags(sot)
        self.assertEqual(["blue"], tags)

        self.user_cloud.network.remove_tag(sot, "blue")
        tags = self.user_cloud.network.get_tags(sot)
        self.assertEqual([], tags)

        # The operation is not idempotent.
        self.assertRaises(
            exceptions.NotFoundException,
            self.user_cloud.network.remove_tag,
            sot,
            "blue",
        )

    def test_remove_all_tags(self):
        sot = self.get_command(self.ID)
        self.assertEqual([], sot.tags)

        self.user_cloud.network.set_tags(sot, ["blue", "red"])
        sot = self.get_command(self.ID)
        self.assertEqual(["blue", "red"], sot.tags)

        self.user_cloud.network.remove_all_tags(sot)
        sot = self.get_command(self.ID)
        self.assertEqual([], sot.tags)

    def test_add_tags(self):
        # Skip the test if tag-creation extension is not enabled.
        if not self.user_cloud.network.find_extension("tag-creation"):
            self.skipTest("Network tag-creation extension disabled")

        sot = self.get_command(self.ID)
        self.assertEqual([], sot.tags)

        self.user_cloud.network.add_tags(sot, ["red", "green"])
        self.user_cloud.network.add_tags(sot, ["blue", "yellow"])
        sot = self.get_command(self.ID)
        self.assertEqual(["blue", "green", "red", "yellow"], sot.tags)

        # The operation is idempotent.
        self.user_cloud.network.add_tags(sot, ["blue", "yellow"])
        sot = self.get_command(self.ID)
        self.assertEqual(["blue", "green", "red", "yellow"], sot.tags)

        self.user_cloud.network.add_tags(sot, [])
        sot = self.get_command(self.ID)
        self.assertEqual(["blue", "green", "red", "yellow"], sot.tags)



# Copyright 2015 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import time

from openstack.cloud import meta
from openstack import exceptions


# TODO(stephenfin): Convert to use real resources
def get_events(cloud, stack_id, event_args, marker=None, limit=None):
    # TODO(mordred) FIX THIS ONCE assert_calls CAN HANDLE QUERY STRINGS
    params = collections.OrderedDict()
    for k in sorted(event_args.keys()):
        params[k] = event_args[k]

    if marker:
        event_args['marker'] = marker
    if limit:
        event_args['limit'] = limit

    response = cloud.orchestration.get(
        f'/stacks/{stack_id}/events',
        params=params,
    )
    exceptions.raise_from_response(response)

    # Show which stack the event comes from (for nested events)
    events = meta.get_and_munchify('events', response.json())
    for e in events:
        e['stack_name'] = stack_id.split("/")[0]
    return events


def poll_for_events(
    cloud, stack_name, action=None, poll_period=5, marker=None
):
    """Continuously poll events and logs for performed action on stack."""

    def stop_check_action(a):
        stop_status = (f'{action}_FAILED', f'{action}_COMPLETE')
        return a in stop_status

    def stop_check_no_action(a):
        return a.endswith('_COMPLETE') or a.endswith('_FAILED')

    if action:
        stop_check = stop_check_action
    else:
        stop_check = stop_check_no_action

    no_event_polls = 0
    msg_template = "\n Stack %(name)s %(status)s \n"

    def is_stack_event(event):
        if (
            event.get('resource_name', '') != stack_name
            and event.get('physical_resource_id', '') != stack_name
        ):
            return False

        phys_id = event.get('physical_resource_id', '')
        links = {
            link.get('rel'): link.get('href')
            for link in event.get('links', [])
        }
        stack_id = links.get('stack', phys_id).rsplit('/', 1)[-1]
        return stack_id == phys_id

    while True:
        events = get_events(
            cloud,
            stack_id=stack_name,
            event_args={'sort_dir': 'asc', 'marker': marker},
        )

        if len(events) == 0:
            no_event_polls += 1
        else:
            no_event_polls = 0
            # set marker to last event that was received.
            marker = getattr(events[-1], 'id', None)

            for event in events:
                # check if stack event was also received
                if is_stack_event(event):
                    stack_status = getattr(event, 'resource_status', '')
                    msg = msg_template % dict(
                        name=stack_name, status=stack_status
                    )
                    if stop_check(stack_status):
                        return stack_status, msg

        if no_event_polls >= 2:
            # after 2 polls with no events, fall back to a stack get
            stack = cloud.get_stack(stack_name, resolve_outputs=False)
            if stack:
                stack_status = stack['stack_status']
                msg = msg_template % dict(name=stack_name, status=stack_status)
                if stop_check(stack_status):
                    return stack_status, msg
                # go back to event polling again
                no_event_polls = 0

        time.sleep(poll_period)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.shared_file_system.v2 import user_message
from openstack.tests.unit import base


IDENTIFIER = "2784bc88-b729-4220-a6bb-a8b7a8f53aad"
EXAMPLE = {
    "id": IDENTIFIER,
    "project_id": "dcc9de3c5fc8471ba3662dbb2b6166d5",
    "action_id": "001",
    "detail_id": "008",
    "message_level": "ERROR",
    "created_at": "2021-03-26T05:16:39.000000",
    "expires_at": "2021-04-25T05:16:39.000000",
    "request_id": "req-e4b3e6de-ce4d-4ef2-b1e7-0087200e4db3",
    "resource_type": "SHARE",
    "resource_id": "c2e4ca07-8c37-4014-92c9-2171c7813fa0",
    "user_message": (
        "allocate host: No storage could be allocated"
        "for this share request, Capabilities filter"
        "didn't succeed."
    ),
}


class TestUserMessage(base.TestCase):
    def test_basic(self):
        message = user_message.UserMessage()
        self.assertEqual('messages', message.resources_key)
        self.assertEqual('/messages', message.base_path)
        self.assertTrue(message.allow_list)
        self.assertFalse(message.allow_create)
        self.assertFalse(message.allow_commit)
        self.assertTrue(message.allow_delete)
        self.assertTrue(message.allow_fetch)
        self.assertFalse(message.allow_head)

        self.assertDictEqual(
            {"limit": "limit", "marker": "marker", "message_id": "message_id"},
            message._query_mapping._mapping,
        )

    def test_user_message(self):
        messages = user_message.UserMessage(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], messages.id)
        self.assertEqual(EXAMPLE['resource_id'], messages.resource_id)
        self.assertEqual(EXAMPLE['message_level'], messages.message_level)
        self.assertEqual(EXAMPLE['user_message'], messages.user_message)
        self.assertEqual(EXAMPLE['expires_at'], messages.expires_at)
        self.assertEqual(EXAMPLE['detail_id'], messages.detail_id)
        self.assertEqual(EXAMPLE['created_at'], messages.created_at)
        self.assertEqual(EXAMPLE['request_id'], messages.request_id)
        self.assertEqual(EXAMPLE['project_id'], messages.project_id)
        self.assertEqual(EXAMPLE['resource_type'], messages.resource_type)
        self.assertEqual(EXAMPLE['action_id'], messages.action_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.dns.v2 import tld as _tld
from openstack.tests.functional import base


class TestTLD(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()

        self.require_service('dns')

        self.tld_name = 'xyz'
        self.tld_description = 'The xyz TLD'

    def test_tld(self):
        # create the tld

        tld = self.operator_cloud.dns.create_tld(
            name=self.tld_name, description=self.tld_description
        )
        self.assertIsInstance(tld, _tld.TLD)
        self.assertEqual(self.tld_description, tld.description)
        self.addCleanup(self.operator_cloud.dns.delete_tld, tld)

        # update the tld

        tld = self.operator_cloud.dns.update_tld(
            tld, description=self.tld_description
        )
        self.assertIsInstance(tld, _tld.TLD)
        self.assertEqual(self.tld_description, tld.description)

        # retrieve details of the (updated) tld by ID

        tld = self.operator_cloud.dns.get_tld(tld.id)
        self.assertIsInstance(tld, _tld.TLD)
        self.assertEqual(self.tld_description, tld.description)

        # retrieve details of the (updated) tld by name

        tld = self.operator_cloud.dns.find_tld(tld.name)
        self.assertIsInstance(tld, _tld.TLD)
        self.assertEqual(self.tld_description, tld.description)

        # list all tlds
        tlds = list(self.operator_cloud.dns.tlds())
        self.assertIsInstance(tlds[0], _tld.TLD)
        self.assertIn(
            self.tld_name, {x.name for x in self.operator_cloud.dns.tlds()}
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.shared_file_system.v2 import resource_locks as _resource_locks
from openstack.tests.functional.shared_file_system import base


class ResourceLocksTest(base.BaseSharedFileSystemTest):
    def setUp(self):
        super().setUp()

        self.SHARE_NAME = self.getUniqueString()
        share = self.user_cloud.shared_file_system.create_share(
            name=self.SHARE_NAME,
            size=2,
            share_type="dhss_false",
            share_protocol='NFS',
            description=None,
        )
        self.SHARE_ID = share.id
        self.user_cloud.shared_file_system.wait_for_status(
            share,
            status='available',
            failures=['error'],
            interval=5,
            wait=self._wait_for_timeout,
        )
        access_rule = self.user_cloud.share.create_access_rule(
            self.SHARE_ID,
            access_level="rw",
            access_type="ip",
            access_to="0.0.0.0/0",
        )
        self.user_cloud.shared_file_system.wait_for_status(
            access_rule,
            status='active',
            failures=['error'],
            interval=5,
            wait=self._wait_for_timeout,
            attribute='state',
        )
        self.assertIsNotNone(share)
        self.assertIsNotNone(share.id)
        self.ACCESS_ID = access_rule.id
        share_lock = self.create_resource_lock(
            resource_action='delete',
            resource_type='share',
            resource_id=self.SHARE_ID,
            lock_reason='openstacksdk testing',
        )
        access_lock = self.create_resource_lock(
            resource_action='show',
            resource_type='access_rule',
            resource_id=self.ACCESS_ID,
            lock_reason='openstacksdk testing',
        )
        self.SHARE_LOCK_ID = share_lock.id
        self.ACCESS_LOCK_ID = access_lock.id

    def test_get(self):
        share_lock = self.user_cloud.shared_file_system.get_resource_lock(
            self.SHARE_LOCK_ID
        )
        access_lock = self.user_cloud.shared_file_system.get_resource_lock(
            self.ACCESS_LOCK_ID
        )
        assert isinstance(share_lock, _resource_locks.ResourceLock)
        assert isinstance(access_lock, _resource_locks.ResourceLock)
        self.assertEqual(self.SHARE_LOCK_ID, share_lock.id)
        self.assertEqual(self.ACCESS_LOCK_ID, access_lock.id)
        self.assertEqual('show', access_lock.resource_action)

    def test_list(self):
        resource_locks = self.user_cloud.share.resource_locks()
        self.assertGreater(len(list(resource_locks)), 0)
        lock_attrs = (
            'id',
            'lock_reason',
            'resource_type',
            'resource_action',
            'lock_context',
            'created_at',
            'updated_at',
        )
        for lock in resource_locks:
            for attribute in lock_attrs:
                self.assertTrue(hasattr(lock, attribute))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack.compute.v2 import server_action
from openstack.tests.unit import base

EXAMPLE = {
    'action': 'stop',
    'events': [
        {
            'event': 'compute_stop_instance',
            'finish_time': '2018-04-25T01:26:36.790544',
            'host': 'compute',
            'hostId': '2091634baaccdc4c5a1d57069c833e402921df696b7f970791b12ec6',  # noqa: E501
            'result': 'Success',
            'start_time': '2018-04-25T01:26:36.539271',
            'traceback': None,
            'details': None,
        }
    ],
    'instance_uuid': '4bf3473b-d550-4b65-9409-292d44ab14a2',
    'message': None,
    'project_id': '6f70656e737461636b20342065766572',
    'request_id': 'req-0d819d5c-1527-4669-bdf0-ffad31b5105b',
    'start_time': '2018-04-25T01:26:36.341290',
    'updated_at': '2018-04-25T01:26:36.790544',
    'user_id': 'admin',
}


class TestServerAction(base.TestCase):
    def setUp(self):
        super().setUp()
        self.resp = mock.Mock()
        self.resp.body = None
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.resp.status_code = 200
        self.sess = mock.Mock()
        self.sess.post = mock.Mock(return_value=self.resp)

    def test_basic(self):
        sot = server_action.ServerAction()
        self.assertEqual('instanceAction', sot.resource_key)
        self.assertEqual('instanceActions', sot.resources_key)
        self.assertEqual(
            '/servers/%(server_id)s/os-instance-actions',
            sot.base_path,
        )
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_list)
        self.assertFalse(sot.allow_create)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)

        self.assertDictEqual(
            {
                'changes_before': 'changes-before',
                'changes_since': 'changes-since',
                'limit': 'limit',
                'marker': 'marker',
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = server_action.ServerAction(**EXAMPLE)
        self.assertEqual(EXAMPLE['action'], sot.action)
        # FIXME: This isn't populated since it conflicts with the server_id URI
        # argument
        # self.assertEqual(EXAMPLE['instance_uuid'], sot.server_id)
        self.assertEqual(EXAMPLE['message'], sot.message)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)
        self.assertEqual(EXAMPLE['request_id'], sot.request_id)
        self.assertEqual(EXAMPLE['start_time'], sot.start_time)
        self.assertEqual(EXAMPLE['user_id'], sot.user_id)
        self.assertEqual(
            [server_action.ServerActionEvent(**e) for e in EXAMPLE['events']],
            sot.events,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.baremetal.v1 import _common
from openstack import resource


class VolumeConnector(resource.Resource):
    resources_key = 'connectors'
    base_path = '/volume/connectors'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    allow_patch = True
    commit_method = 'PATCH'
    commit_jsonpatch = True

    _query_mapping = resource.QueryParameters(
        'node',
        'detail',
        fields={'type': _common.fields_type},
    )

    # Volume Connectors is available since 1.32
    _max_microversion = '1.32'

    #: The identifier of Volume connector and this field depends on the "type"
    # of the volume_connector
    connector_id = resource.Body('connector_id')
    #: Timestamp at which the port was created.
    created_at = resource.Body('created_at')
    #: A set of one or more arbitrary metadata key and value pairs.
    extra = resource.Body('extra')
    #: A list of relative links, including the self and bookmark links.
    links = resource.Body('links', type=list)
    #: The UUID of node this port belongs to
    node_id = resource.Body('node_uuid')
    #: The types of Volume connector
    type = resource.Body('type')
    #: Timestamp at which the port was last updated.
    updated_at = resource.Body('updated_at')
    #: The UUID of the port
    id = resource.Body('uuid', alternate_id=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# mypy: disable-error-code="method-assign"

from openstack.network.v2 import subnet_pool as _subnet_pool
from openstack.tests.functional.network.v2 import common


class TestSubnetPool(common.TestTagNeutron):
    SUBNET_POOL_ID = None
    MINIMUM_PREFIX_LENGTH = 8
    DEFAULT_PREFIX_LENGTH = 24
    MAXIMUM_PREFIX_LENGTH = 32
    DEFAULT_QUOTA = 24
    IS_SHARED = False
    IP_VERSION = 4
    PREFIXES = ["10.100.0.0/24", "10.101.0.0/24"]

    def setUp(self):
        super().setUp()
        self.SUBNET_POOL_NAME = self.getUniqueString()
        self.SUBNET_POOL_NAME_UPDATED = self.getUniqueString()
        subnet_pool = self.user_cloud.network.create_subnet_pool(
            name=self.SUBNET_POOL_NAME,
            min_prefixlen=self.MINIMUM_PREFIX_LENGTH,
            default_prefixlen=self.DEFAULT_PREFIX_LENGTH,
            max_prefixlen=self.MAXIMUM_PREFIX_LENGTH,
            default_quota=self.DEFAULT_QUOTA,
            shared=self.IS_SHARED,
            prefixes=self.PREFIXES,
        )
        assert isinstance(subnet_pool, _subnet_pool.SubnetPool)
        self.assertEqual(self.SUBNET_POOL_NAME, subnet_pool.name)
        self.SUBNET_POOL_ID = self.ID = subnet_pool.id
        self.get_command = self.user_cloud.network.get_subnet_pool

    def tearDown(self):
        sot = self.user_cloud.network.delete_subnet_pool(self.SUBNET_POOL_ID)
        self.assertIsNone(sot)
        super().tearDown()

    def test_find(self):
        sot = self.user_cloud.network.find_subnet_pool(self.SUBNET_POOL_NAME)
        self.assertEqual(self.SUBNET_POOL_ID, sot.id)

    def test_get(self):
        sot = self.user_cloud.network.get_subnet_pool(self.SUBNET_POOL_ID)
        self.assertEqual(self.SUBNET_POOL_NAME, sot.name)
        self.assertEqual(self.MINIMUM_PREFIX_LENGTH, sot.minimum_prefix_length)
        self.assertEqual(self.DEFAULT_PREFIX_LENGTH, sot.default_prefix_length)
        self.assertEqual(self.MAXIMUM_PREFIX_LENGTH, sot.maximum_prefix_length)
        self.assertEqual(self.DEFAULT_QUOTA, sot.default_quota)
        self.assertEqual(self.IS_SHARED, sot.is_shared)
        self.assertEqual(self.IP_VERSION, sot.ip_version)
        self.assertEqual(self.PREFIXES, sot.prefixes)

    def test_list(self):
        names = [o.name for o in self.user_cloud.network.subnet_pools()]
        self.assertIn(self.SUBNET_POOL_NAME, names)

    def test_update(self):
        sot = self.user_cloud.network.update_subnet_pool(
            self.SUBNET_POOL_ID, name=self.SUBNET_POOL_NAME_UPDATED
        )
        self.assertEqual(self.SUBNET_POOL_NAME_UPDATED, sot.name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from openstack.message.v2 import claim as _claim
from openstack.message.v2 import message as _message
from openstack.message.v2 import queue as _queue
from openstack.message.v2 import subscription as _subscription
from openstack import proxy
from openstack import resource


class Proxy(proxy.Proxy):
    _resource_registry = {
        "claim": _claim.Claim,
        "message": _message.Message,
        "queue": _queue.Queue,
        "subscription": _subscription.Subscription,
    }

    def create_queue(self, **attrs):
        """Create a new queue from attributes

        :param dict attrs: Keyword arguments which will be used to create
            a :class:`~openstack.message.v2.queue.Queue`,
            comprised of the properties on the Queue class.

        :returns: The results of queue creation
        :rtype: :class:`~openstack.message.v2.queue.Queue`
        """
        return self._create(_queue.Queue, **attrs)

    def get_queue(self, queue):
        """Get a queue

        :param queue: The value can be the name of a queue or a
            :class:`~openstack.message.v2.queue.Queue` instance.

        :returns: One :class:`~openstack.message.v2.queue.Queue`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            queue matching the name could be found.
        """
        return self._get(_queue.Queue, queue)

    def queues(self, **query):
        """Retrieve a generator of queues

        :param kwargs query: Optional query parameters to be sent to
            restrict the queues to be returned. Available parameters include:

            * limit: Requests at most the specified number of items be
                returned from the query.
            * marker: Specifies the ID of the last-seen queue. Use the limit
                parameter to make an initial limited request and use the ID of
                the last-seen queue from the response as the marker parameter
                value in a subsequent limited request.

        :returns: A generator of queue instances.
        """
        return self._list(_queue.Queue, **query)

    def delete_queue(self, value, ignore_missing=True):
        """Delete a queue

        :param value: The value can be either the name of a queue or a
            :class:`~openstack.message.v2.queue.Queue` instance.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the queue does not exist.
            When set to ``True``, no exception will be set when
            attempting to delete a nonexistent queue.

        :returns: ``None``
        """
        return self._delete(_queue.Queue, value, ignore_missing=ignore_missing)

    def post_message(self, queue_name, messages):
        """Post messages to given queue

        :param queue_name: The name of target queue to post message to.
        :param messages: List of messages body and TTL to post.
            :type messages: :py:class:`list`

        :returns: A string includes location of messages successfully posted.
        """
        message = self._get_resource(
            _message.Message, None, queue_name=queue_name
        )
        return message.post(self, messages)

    def messages(self, queue_name, **query):
        """Retrieve a generator of messages

        :param queue_name: The name of target queue to query messages from.
        :param kwargs query: Optional query parameters to be sent to
            restrict the messages to be returned. Available parameters include:

            * limit: Requests at most the specified number of items be
                returned from the query.
            * marker: Specifies the ID of the last-seen subscription. Use the
                limit parameter to make an initial limited request and use the
                ID of the last-seen subscription from the response as the
                marker parameter value in a subsequent limited request.
            * echo: Indicate if the messages can be echoed back to the client
                that posted them.
            * include_claimed: Indicate if the messages list should include
                the claimed messages.

        :returns: A generator of message instances.
        """
        query["queue_name"] = queue_name
        return self._list(_message.Message, **query)

    def get_message(self, queue_name, message):
        """Get a message

        :param queue_name: The name of target queue to get message from.
        :param message: The value can be the name of a message or a
            :class:`~openstack.message.v2.message.Message` instance.

        :returns: One :class:`~openstack.message.v2.message.Message`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            message matching the criteria could be found.
        """
        message = self._get_resource(
            _message.Message, message, queue_name=queue_name
        )
        return self._get(_message.Message, message)

    def delete_message(
        self, queue_name, value, claim=None, ignore_missing=True
    ):
        """Delete a message

        :param queue_name: The name of target queue to delete message from.
        :param value: The value can be either the name of a message or a
            :class:`~openstack.message.v2.message.Message` instance.
        :param claim: The value can be the ID or a
            :class:`~openstack.message.v2.claim.Claim` instance of
            the claim seizing the message. If None, the message has
            not been claimed.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the message does not exist.
            When set to ``True``, no exception will be set when
            attempting to delete a nonexistent message.

        :returns: ``None``
        """
        message = self._get_resource(
            _message.Message, value, queue_name=queue_name
        )
        message.claim_id = resource.Resource._get_id(claim)
        return self._delete(
            _message.Message, message, ignore_missing=ignore_missing
        )

    def create_subscription(self, queue_name, **attrs):
        """Create a new subscription from attributes

        :param queue_name: The name of target queue to subscribe on.
        :param dict attrs: Keyword arguments which will be used to create a
            :class:`~openstack.message.v2.subscription.Subscription`,
            comprised of the properties on the Subscription class.

        :returns: The results of subscription creation
        :rtype: :class:`~openstack.message.v2.subscription.Subscription`
        """
        return self._create(
            _subscription.Subscription, queue_name=queue_name, **attrs
        )

    def subscriptions(self, queue_name, **query):
        """Retrieve a generator of subscriptions

        :param queue_name: The name of target queue to subscribe on.
        :param kwargs query: Optional query parameters to be sent to
            restrict the subscriptions to be returned. Available parameters
            include:

            * limit: Requests at most the specified number of items be
                returned from the query.
            * marker: Specifies the ID of the last-seen subscription. Use the
                limit parameter to make an initial limited request and use the
                ID of the last-seen subscription from the response as the
                marker parameter value in a subsequent limited request.

        :returns: A generator of subscription instances.
        """
        query["queue_name"] = queue_name
        return self._list(_subscription.Subscription, **query)

    def get_subscription(self, queue_name, subscription):
        """Get a subscription

        :param queue_name: The name of target queue of subscription.
        :param message: The value can be the ID of a subscription or a
            :class:`~openstack.message.v2.subscription.Subscription` instance.

        :returns: One :class:`~openstack.message.v2.subscription.Subscription`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            subscription matching the criteria could be found.
        """
        subscription = self._get_resource(
            _subscription.Subscription, subscription, queue_name=queue_name
        )
        return self._get(_subscription.Subscription, subscription)

    def delete_subscription(self, queue_name, value, ignore_missing=True):
        """Delete a subscription

        :param queue_name: The name of target queue to delete subscription
            from.
        :param value: The value can be either the name of a subscription or a
            :class:`~openstack.message.v2.subscription.Subscription`
            instance.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the subscription does not exist.
            When set to ``True``, no exception will be thrown when
            attempting to delete a nonexistent subscription.

        :returns: ``None``
        """
        subscription = self._get_resource(
            _subscription.Subscription, value, queue_name=queue_name
        )
        return self._delete(
            _subscription.Subscription,
            subscription,
            ignore_missing=ignore_missing,
        )

    def create_claim(self, queue_name, **attrs):
        """Create a new claim from attributes

        :param queue_name: The name of target queue to claim message from.
        :param dict attrs: Keyword arguments which will be used to create a
            :class:`~openstack.message.v2.claim.Claim`,
            comprised of the properties on the Claim class.

        :returns: The results of claim creation
        :rtype: :class:`~openstack.message.v2.claim.Claim`
        """
        return self._create(_claim.Claim, queue_name=queue_name, **attrs)

    def get_claim(self, queue_name, claim):
        """Get a claim

        :param queue_name: The name of target queue to claim message from.
        :param claim: The value can be either the ID of a claim or a
            :class:`~openstack.message.v2.claim.Claim` instance.

        :returns: One :class:`~openstack.message.v2.claim.Claim`
        :raises: :class:`~openstack.exceptions.NotFoundException` when no
            claim matching the criteria could be found.
        """
        return self._get(_claim.Claim, claim, queue_name=queue_name)

    def update_claim(self, queue_name, claim, **attrs):
        """Update an existing claim from attributes

        :param queue_name: The name of target queue to claim message from.
        :param claim: The value can be either the ID of a claim or a
            :class:`~openstack.message.v2.claim.Claim` instance.
        :param dict attrs: Keyword arguments which will be used to update a
            :class:`~openstack.message.v2.claim.Claim`,
            comprised of the properties on the Claim class.

        :returns: The results of claim update
        :rtype: :class:`~openstack.message.v2.claim.Claim`
        """
        return self._update(
            _claim.Claim, claim, queue_name=queue_name, **attrs
        )

    def delete_claim(self, queue_name, claim, ignore_missing=True):
        """Delete a claim

        :param queue_name: The name of target queue to claim messages from.
        :param claim: The value can be either the ID of a claim or a
            :class:`~openstack.message.v2.claim.Claim` instance.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be
            raised when the claim does not exist.
            When set to ``True``, no exception will be thrown when
            attempting to delete a nonexistent claim.

        :returns: ``None``
        """
        return self._delete(
            _claim.Claim,
            claim,
            queue_name=queue_name,
            ignore_missing=ignore_missing,
        )

    # ========== Utilities ==========

    def wait_for_status(
        self,
        res: resource.ResourceT,
        status: str,
        failures: list[str] | None = None,
        interval: int | float | None = 2,
        wait: int | None = None,
        attribute: str = 'status',
        callback: ty.Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for the resource to be in a particular status.

        :param session: The session to use for making this request.
        :param resource: The resource to wait on to reach the status. The
            resource must have a status attribute specified via ``attribute``.
        :param status: Desired status of the resource.
        :param failures: Statuses that would indicate the transition
            failed such as 'ERROR'. Defaults to ['ERROR'].
        :param interval: Number of seconds to wait between checks.
        :param wait: Maximum number of seconds to wait for transition.
            Set to ``None`` to wait forever.
        :param attribute: Name of the resource attribute that contains the
            status.
        :param callback: A callback function. This will be called with a single
            value, progress. This is API specific but is generally a percentage
            value from 0-100.

        :return: The updated resource.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if the
            transition to status failed to occur in ``wait`` seconds.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
            transitioned to one of the states in ``failures``.
        :raises: :class:`~AttributeError` if the resource does not have a
            ``status`` attribute
        """
        return resource.wait_for_status(
            self, res, status, failures, interval, wait, attribute, callback
        )

    def wait_for_delete(
        self,
        res: resource.ResourceT,
        interval: int = 2,
        wait: int = 120,
        callback: ty.Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for a resource to be deleted.

        :param res: The resource to wait on to be deleted.
        :param interval: Number of seconds to wait before to consecutive
            checks.
        :param wait: Maximum number of seconds to wait before the change.
        :param callback: A callback function. This will be called with a single
            value, progress, which is a percentage value from 0-100.

        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
            to delete failed to occur in the specified seconds.
        """
        return resource.wait_for_delete(self, res, interval, wait, callback)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import warnings

from openstack import exceptions
from openstack import resource
from openstack import utils
from openstack import warnings as os_warnings


class GroupSnapshot(resource.Resource):
    resource_key = "group_snapshot"
    resources_key = "group_snapshots"
    base_path = "/group_snapshots"

    # capabilities
    allow_fetch = True
    allow_create = True
    allow_delete = True
    allow_commit = False
    allow_list = True

    _query_mapping = resource.QueryParameters(
        "limit",
        "marker",
        "offset",
        "sort_dir",
        "sort_key",
        "sort",
        all_projects="all_tenants",
    )

    #: Properties
    #: The date and time when the resource was created.
    created_at = resource.Body("created_at")
    #: The group snapshot description.
    description = resource.Body("description")
    #: The UUID of the source group.
    group_id = resource.Body("group_id")
    #: The group type ID.
    group_type_id = resource.Body("group_type_id")
    #: The ID of the group snapshot.
    id = resource.Body("id")
    #: The group snapshot name.
    name = resource.Body("name")
    #: The UUID of the volume group snapshot project.
    project_id = resource.Body("project_id")
    #: The status of the generic group snapshot.
    status = resource.Body("status")

    # Pagination support was added in microversion 3.29
    _max_microversion = '3.29'

    def _action(self, session, body, microversion=None):
        """Preform aggregate actions given the message body."""
        url = utils.urljoin(self.base_path, self.id, 'action')
        headers = {'Accept': ''}
        # TODO(stephenfin): This logic belongs in openstack.resource I suspect
        if microversion is None:
            if session.default_microversion:
                microversion = session.default_microversion
            else:
                microversion = utils.maximum_supported_microversion(
                    session,
                    self._max_microversion,
                )
        response = session.post(
            url,
            json=body,
            headers=headers,
            microversion=microversion,
        )
        exceptions.raise_from_response(response)
        return response

    def reset_status(self, session, state):
        """Resets the status for a group snapshot."""
        body = {'reset_status': {'status': state}}
        return self._action(session, body)

    def reset_state(self, session, status):
        warnings.warn(
            "reset_state is a deprecated alias for reset_status and will be "
            "removed in a future release.",
            os_warnings.RemovedInSDK60Warning,
        )
        self.reset_status(session, status)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.shared_file_system.v2 import share_group
from openstack.tests.unit import base


EXAMPLE = {
    "status": "creating",
    "description": None,
    "links": "[]",
    "availability_zone": None,
    "source_share_group_snapshot_id": None,
    "share_network_id": None,
    "share_server_id": None,
    "host": None,
    "share_group_type_id": "89861c2a-10bf-4013-bdd4-3d020466aee4",
    "consistent_snapshot_support": None,
    "id": "f9c1f80c-2392-4e34-bd90-fc89cdc5bf93",
    "name": None,
    "created_at": "2021-06-03T19:20:33.974421",
    "project_id": "e23850eeb91d4fa3866af634223e454c",
    "share_types": ["ecd11f4c-d811-4471-b656-c755c77e02ba"],
}


class TestShareGroups(base.TestCase):
    def test_basic(self):
        share_groups = share_group.ShareGroup()
        self.assertEqual('share_groups', share_groups.resources_key)
        self.assertEqual('/share-groups', share_groups.base_path)
        self.assertTrue(share_groups.allow_list)
        self.assertTrue(share_groups.allow_fetch)
        self.assertTrue(share_groups.allow_create)
        self.assertTrue(share_groups.allow_commit)
        self.assertTrue(share_groups.allow_delete)
        self.assertFalse(share_groups.allow_head)

        self.assertDictEqual(
            {
                "limit": "limit",
                "marker": "marker",
                "share_group_id": "share_group_id",
            },
            share_groups._query_mapping._mapping,
        )

    def test_make_share_groups(self):
        share_group_res = share_group.ShareGroup(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], share_group_res.id)
        self.assertEqual(EXAMPLE['status'], share_group_res.status)
        self.assertEqual(
            EXAMPLE['availability_zone'], share_group_res.availability_zone
        )
        self.assertEqual(EXAMPLE['description'], share_group_res.description)
        self.assertEqual(
            EXAMPLE['source_share_group_snapshot_id'],
            share_group_res.share_group_snapshot_id,
        )
        self.assertEqual(
            EXAMPLE['share_network_id'], share_group_res.share_network_id
        )
        self.assertEqual(
            EXAMPLE['share_group_type_id'], share_group_res.share_group_type_id
        )
        self.assertEqual(
            EXAMPLE['consistent_snapshot_support'],
            share_group_res.consistent_snapshot_support,
        )
        self.assertEqual(EXAMPLE['created_at'], share_group_res.created_at)
        self.assertEqual(EXAMPLE['project_id'], share_group_res.project_id)
        self.assertEqual(EXAMPLE['share_types'], share_group_res.share_types)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from calendar import timegm
import collections
import functools
from hashlib import sha1
import hmac
import json
import os
import time
import typing as ty
from urllib import parse

from openstack import _log
from openstack.cloud import _utils
from openstack import exceptions
from openstack.object_store.v1 import account as _account
from openstack.object_store.v1 import container as _container
from openstack.object_store.v1 import info as _info
from openstack.object_store.v1 import obj as _obj
from openstack import proxy
from openstack import resource
from openstack import utils

DEFAULT_OBJECT_SEGMENT_SIZE = 1073741824  # 1GB
DEFAULT_MAX_FILE_SIZE = (5 * 1024 * 1024 * 1024 + 2) / 2
EXPIRES_ISO8601_FORMAT = '%Y-%m-%dT%H:%M:%SZ'
SHORT_EXPIRES_ISO8601_FORMAT = '%Y-%m-%d'


def _get_expiration(expiration):
    return int(time.time() + expiration)


class Proxy(proxy.Proxy):
    _resource_registry = {
        "account": _account.Account,
        "container": _container.Container,
        "info": _info.Info,
        "object": _obj.Object,
    }

    skip_discovery = True

    Account = _account.Account
    Container = _container.Container
    Object = _obj.Object

    log = _log.setup_logging('openstack')

    @functools.lru_cache(maxsize=256)
    def _extract_name(self, url, service_type=None, project_id=None):
        url_path = parse.urlparse(url).path.strip()
        # Remove / from the beginning to keep the list indexes of interesting
        # things consistent
        if url_path.startswith('/'):
            url_path = url_path[1:]

        # Split url into parts and exclude potential project_id in some urls
        url_parts = [
            x
            for x in url_path.split('/')
            if (
                x != project_id
                and (
                    not project_id
                    or (project_id and x != 'AUTH_' + project_id)
                )
            )
        ]
        # Strip leading version piece so that
        # GET /v1/AUTH_xxx
        # returns ['AUTH_xxx']
        if (
            url_parts[0]
            and url_parts[0][0] == 'v'
            and url_parts[0][1]
            and url_parts[0][1].isdigit()
        ):
            url_parts = url_parts[1:]

        # Strip out anything that's empty or None
        parts = [part for part in url_parts if part]

        # Getting the root of an endpoint is doing version discovery
        if not parts:
            return ['account']

        if len(parts) == 1:
            if 'endpoints' in parts:
                return ['endpoints']
            else:
                return ['container']
        else:
            return ['object']

    def get_account_metadata(self):
        """Get metadata for this account.

        :rtype:
            :class:`~openstack.object_store.v1.account.Account`
        """
        return self._head(_account.Account)

    def set_account_metadata(self, **metadata):
        """Set metadata for this account.

        :param kwargs metadata: Key/value pairs to be set as metadata on the
            container. Custom metadata can be set. Custom metadata are keys and
            values defined by the user.
        """
        account = self._get_resource(_account.Account, None)
        account.set_metadata(self, metadata)

    def delete_account_metadata(self, keys):
        """Delete metadata for this account.

        :param keys: The keys of metadata to be deleted.
        """
        account = self._get_resource(_account.Account, None)
        account.delete_metadata(self, keys)

    def containers(self, **query):
        """Obtain Container objects for this account.

        :param kwargs query: Optional query parameters to be sent to limit
            the resources being returned.

        :rtype: A generator of
            :class:`~openstack.object_store.v1.container.Container` objects.
        """
        return self._list(_container.Container, paginated=True, **query)

    def create_container(self, name, **attrs):
        """Create a new container from attributes

        :param container: Name of the container to create.
        :param dict attrs: Keyword arguments which will be used to create
            a :class:`~openstack.object_store.v1.container.Container`,
            comprised of the properties on the Container class.

        :returns: The results of container creation
        :rtype: :class:`~openstack.object_store.v1.container.Container`
        """
        return self._create(_container.Container, name=name, **attrs)

    def delete_container(self, container, ignore_missing=True):
        """Delete a container

        :param container: The value can be either the name of a container or a
            :class:`~openstack.object_store.v1.container.Container` instance.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the container does not exist. When set to ``True``, no
            exception will be set when attempting to delete a nonexistent
            server.

        :returns: ``None``
        """
        self._delete(
            _container.Container, container, ignore_missing=ignore_missing
        )

    def get_container_metadata(self, container):
        """Get metadata for a container

        :param container: The value can be the name of a container or a
            :class:`~openstack.object_store.v1.container.Container` instance.

        :returns: One :class:`~openstack.object_store.v1.container.Container`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        return self._head(_container.Container, container)

    def set_container_metadata(self, container, refresh=True, **metadata):
        """Set metadata for a container.

        :param container: The value can be the name of a container or a
            :class:`~openstack.object_store.v1.container.Container`
            instance.
        :param refresh: Flag to trigger refresh of container object re-fetch.
        :param kwargs metadata: Key/value pairs to be set as metadata on the
            container. Both custom and system metadata can be set. Custom
            metadata are keys and values defined by the user. System metadata
            are keys defined by the Object Store and values defined by the
            user. The system metadata keys are:

            - `content_type`
            - `is_content_type_detected`
            - `versions_location`
            - `read_ACL`
            - `write_ACL`
            - `sync_to`
            - `sync_key`
        """
        res = self._get_resource(_container.Container, container)
        res.set_metadata(self, metadata, refresh=refresh)
        return res

    def delete_container_metadata(self, container, keys):
        """Delete metadata for a container.

        :param container: The value can be the ID of a container or a
            :class:`~openstack.object_store.v1.container.Container` instance.
        :param keys: The keys of metadata to be deleted.
        """
        res = self._get_resource(_container.Container, container)
        res.delete_metadata(self, keys)
        return res

    def objects(self, container, **query):
        """Return a generator that yields the Container's objects.

        :param container: A container object or the name of a container
            that you want to retrieve objects from.
        :type container:
            :class:`~openstack.object_store.v1.container.Container`
        :param kwargs query: Optional query parameters to be sent to limit
            the resources being returned.

        :rtype: A generator of
            :class:`~openstack.object_store.v1.obj.Object` objects.
        """
        container = self._get_container_name(container=container)

        for obj in self._list(
            _obj.Object,
            container=container,
            paginated=True,
            format='json',
            **query,
        ):
            obj.container = container
            yield obj

    def _get_container_name(self, obj=None, container=None):
        if obj is not None:
            obj = self._get_resource(_obj.Object, obj)
            if obj.container is not None:
                return obj.container
        if container is not None:
            container = self._get_resource(_container.Container, container)
            return container.name

        raise ValueError("container must be specified")

    def get_object(
        self,
        obj,
        container=None,
        resp_chunk_size=1024,
        outfile=None,
        remember_content=False,
    ):
        """Get the data associated with an object

        :param obj: The value can be the name of an object or a
            :class:`~openstack.object_store.v1.obj.Object` instance.
        :param container: The value can be the name of a container or a
            :class:`~openstack.object_store.v1.container.Container` instance.
        :param int resp_chunk_size: chunk size of data to read. Only used if
            the results are being written to a file or stream is True.
            (optional, defaults to 1k)
        :param outfile: Write the object to a file instead of returning the
            contents. If this option is given, body in the return tuple will be
            None. outfile can either be a file path given as a string, or a
            File like object.
        :param bool remember_content: Flag whether object data should be saved
            as `data` property of the Object. When left as `false` and
            `outfile` is not defined data will not be saved and need to be
            fetched separately.

        :returns: Instance of the
            :class:`~openstack.object_store.v1.obj.Object` objects.
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        container_name = self._get_container_name(obj=obj, container=container)

        _object = self._get_resource(
            _obj.Object, obj, container=container_name
        )
        request = _object._prepare_request()

        get_stream = outfile is not None

        response = self.get(
            request.url, headers=request.headers, stream=get_stream
        )
        exceptions.raise_from_response(response)
        _object._translate_response(response, has_body=False)

        if outfile:
            if isinstance(outfile, str):
                outfile_handle = open(outfile, 'wb')
            else:
                outfile_handle = outfile
            for chunk in response.iter_content(
                resp_chunk_size, decode_unicode=False
            ):
                outfile_handle.write(chunk)
            if isinstance(outfile, str):
                outfile_handle.close()
            else:
                outfile_handle.flush()
        elif remember_content:
            _object.data = response.text

        return _object

    def download_object(self, obj, container=None, **attrs):
        """Download the data contained inside an object.

        :param obj: The value can be the name of an object or a
            :class:`~openstack.object_store.v1.obj.Object` instance.
        :param container: The value can be the name of a container or a
            :class:`~openstack.object_store.v1.container.Container` instance.

        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        container_name = self._get_container_name(obj=obj, container=container)
        obj = self._get_resource(
            _obj.Object, obj, container=container_name, **attrs
        )
        return obj.download(self)

    def stream_object(self, obj, container=None, chunk_size=1024, **attrs):
        """Stream the data contained inside an object.

        :param obj: The value can be the name of an object or a
            :class:`~openstack.object_store.v1.obj.Object` instance.
        :param container: The value can be the name of a container or a
            :class:`~openstack.object_store.v1.container.Container` instance.

        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        :returns: An iterator that iterates over chunk_size bytes
        """
        container_name = self._get_container_name(obj=obj, container=container)
        obj = self._get_resource(
            _obj.Object, obj, container=container_name, **attrs
        )
        return obj.stream(self, chunk_size=chunk_size)

    def create_object(
        self,
        container,
        name,
        filename=None,
        md5=None,
        sha256=None,
        segment_size=None,
        use_slo=True,
        metadata=None,
        generate_checksums=None,
        data=None,
        **headers,
    ):
        """Create a file object.

        Automatically uses large-object segments if needed.

        :param container: The name of the container to store the file in.
            This container will be created if it does not exist already.
        :param name: Name for the object within the container.
        :param filename: The path to the local file whose contents will be
            uploaded. Mutually exclusive with data.
        :param data: The content to upload to the object. Mutually exclusive
            with filename.
        :param md5: A hexadecimal md5 of the file. (Optional), if it is known
            and can be passed here, it will save repeating the expensive md5
            process. It is assumed to be accurate.
        :param sha256: A hexadecimal sha256 of the file. (Optional) See md5.
        :param segment_size: Break the uploaded object into segments of this
            many bytes. (Optional) SDK will attempt to discover the maximum
            value for this from the server if it is not specified, or will use
            a reasonable default.
        :param headers: These will be passed through to the object creation
            API as HTTP Headers.
        :param use_slo: If the object is large enough to need to be a Large
            Object, use a static rather than dynamic object. Static Objects
            will delete segment objects when the manifest object is deleted.
            (optional, defaults to True)
        :param generate_checksums: Whether to generate checksums on the client
            side that get added to headers for later prevention of double
            uploads of identical data. (optional, defaults to True)
        :param metadata: This dict will get changed into headers that set
            metadata of the object

        :raises: ``:class:`~openstack.exceptions.SDKException``` on operation
            error.
        """
        if data is not None and filename:
            raise ValueError(
                "Both filename and data given. Please choose one."
            )
        if data is not None and not name:
            raise ValueError("name is a required parameter when data is given")
        if data is not None and generate_checksums:
            raise ValueError(
                "checksums cannot be generated with data parameter"
            )
        if generate_checksums is None:
            if data is not None:
                generate_checksums = False
            else:
                generate_checksums = True

        if not metadata:
            metadata = {}

        if not filename and data is None:
            filename = name

        if generate_checksums and (md5 is None or sha256 is None):
            (md5, sha256) = utils._get_file_hashes(filename)
        if md5:
            metadata[self._connection._OBJECT_MD5_KEY] = md5
        if sha256:
            metadata[self._connection._OBJECT_SHA256_KEY] = sha256

        container_name = self._get_container_name(container=container)
        endpoint = f'{container_name}/{name}'

        if data is not None:
            self.log.debug(
                "swift uploading data to %(endpoint)s", {'endpoint': endpoint}
            )
            return self._create(
                _obj.Object,
                container=container_name,
                name=name,
                data=data,
                metadata=metadata,
                **headers,
            )

        # segment_size gets used as a step value in a range call, so needs
        # to be an int
        if segment_size:
            segment_size = int(segment_size)
        segment_size = self.get_object_segment_size(segment_size)
        file_size = os.path.getsize(filename)

        if self.is_object_stale(container_name, name, filename, md5, sha256):
            self._connection.log.debug(
                "swift uploading %(filename)s to %(endpoint)s",
                {'filename': filename, 'endpoint': endpoint},
            )

            if metadata is not None:
                # Rely on the class headers calculation for requested metadata
                meta_headers = _obj.Object()._calculate_headers(metadata)
                headers.update(meta_headers)

            if file_size <= segment_size:
                self._upload_object(endpoint, filename, headers)

            else:
                self._upload_large_object(
                    endpoint,
                    filename,
                    headers,
                    file_size,
                    segment_size,
                    use_slo,
                )

    # Backwards compat
    upload_object = create_object

    def copy_object(self):
        """Copy an object."""
        raise NotImplementedError

    def delete_object(self, obj, ignore_missing=True, container=None):
        """Delete an object

        :param obj: The value can be either the name of an object or a
            :class:`~openstack.object_store.v1.container.Container` instance.
        :param container: The value can be the ID of a container or a
            :class:`~openstack.object_store.v1.container.Container` instance.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.NotFoundException` will be raised
            when the object does not exist.  When set to ``True``, no exception
            will be set when attempting to delete a nonexistent server.

        :returns: ``None``
        """
        container_name = self._get_container_name(obj, container)

        self._delete(
            _obj.Object,
            obj,
            ignore_missing=ignore_missing,
            container=container_name,
        )

    def get_object_metadata(self, obj, container=None):
        """Get metadata for an object.

        :param obj: The value can be the name of an object or a
            :class:`~openstack.object_store.v1.obj.Object` instance.
        :param container: The value can be the ID of a container or a
            :class:`~openstack.object_store.v1.container.Container` instance.

        :returns: One :class:`~openstack.object_store.v1.obj.Object`
        :raises: :class:`~openstack.exceptions.NotFoundException`
            when no resource can be found.
        """
        container_name = self._get_container_name(obj, container)

        return self._head(_obj.Object, obj, container=container_name)

    def set_object_metadata(self, obj, container=None, **metadata):
        """Set metadata for an object.

        Note: This method will do an extra HEAD call.

        :param obj: The value can be the name of an object or a
            :class:`~openstack.object_store.v1.obj.Object` instance.
        :param container: The value can be the name of a container or a
            :class:`~openstack.object_store.v1.container.Container` instance.
        :param kwargs metadata: Key/value pairs to be set as metadata
            on the container. Both custom and system metadata can be set.
            Custom metadata are keys and values defined by the user. System
            metadata are keys defined by the Object Store and values defined by
            the user. The system metadata keys are:

            - `content_type`
            - `content_encoding`
            - `content_disposition`
            - `delete_after`
            - `delete_at`
            - `is_content_type_detected`
        """
        container_name = self._get_container_name(obj, container)
        res = self._get_resource(_obj.Object, obj, container=container_name)
        res.set_metadata(self, metadata)
        return res

    def delete_object_metadata(self, obj, container=None, keys=None):
        """Delete metadata for an object.

        :param obj: The value can be the name of an object or a
            :class:`~openstack.object_store.v1.obj.Object` instance.
        :param container: The value can be the ID of a container or a
            :class:`~openstack.object_store.v1.container.Container` instance.
        :param keys: The keys of metadata to be deleted.
        """
        container_name = self._get_container_name(obj, container)
        res = self._get_resource(_obj.Object, obj, container=container_name)
        res.delete_metadata(self, keys)
        return res

    def is_object_stale(
        self, container, name, filename, file_md5=None, file_sha256=None
    ):
        """Check to see if an object matches the hashes of a file.

        :param container: Name of the container.
        :param name: Name of the object.
        :param filename: Path to the file.
        :param file_md5: Pre-calculated md5 of the file contents. Defaults to
            None which means calculate locally.
        :param file_sha256: Pre-calculated sha256 of the file contents.
            Defaults to None which means calculate locally.
        """
        try:
            metadata = self.get_object_metadata(name, container).metadata
        except exceptions.NotFoundException:
            self._connection.log.debug(
                f"swift stale check, no object: {container}/{name}"
            )
            return True

        if not (file_md5 or file_sha256):
            (file_md5, file_sha256) = utils._get_file_hashes(filename)
        md5_key = metadata.get(
            self._connection._OBJECT_MD5_KEY,
            metadata.get(self._connection._SHADE_OBJECT_MD5_KEY, ''),
        )
        sha256_key = metadata.get(
            self._connection._OBJECT_SHA256_KEY,
            metadata.get(self._connection._SHADE_OBJECT_SHA256_KEY, ''),
        )
        up_to_date = utils._hashes_up_to_date(
            md5=file_md5,
            sha256=file_sha256,
            md5_key=md5_key,
            sha256_key=sha256_key,
        )

        if not up_to_date:
            self._connection.log.debug(
                "swift checksum mismatch: "
                "%(filename)s!=%(container)s/%(name)s",
                {'filename': filename, 'container': container, 'name': name},
            )
            return True

        self._connection.log.debug(
            "swift object up to date: %(container)s/%(name)s",
            {'container': container, 'name': name},
        )
        return False

    def _upload_large_object(
        self, endpoint, filename, headers, file_size, segment_size, use_slo
    ):
        # If the object is big, we need to break it up into segments that
        # are no larger than segment_size, upload each of them individually
        # and then upload a manifest object. The segments can be uploaded in
        # parallel, so we'll use the async feature of the TaskManager.

        segment_futures = []
        segment_results = []
        retry_results = []
        retry_futures = []
        manifest = []

        # Get an OrderedDict with keys being the swift location for the
        # segment, the value a FileSegment file-like object that is a
        # slice of the data for the segment.
        segments = self._get_file_segments(
            endpoint, filename, file_size, segment_size
        )

        # Schedule the segments for upload
        for name, segment in segments.items():
            # Async call to put - schedules execution and returns a future
            segment_future = self._connection._pool_executor.submit(
                self.put, name, headers=headers, data=segment, raise_exc=False
            )
            segment_futures.append(segment_future)
            # TODO(mordred) Collect etags from results to add to this manifest
            # dict. Then sort the list of dicts by path.
            manifest.append(
                dict(
                    # While Object Storage usually expects the name to be
                    # urlencoded in most requests, the SLO manifest requires
                    # plain object names instead.
                    path=f'/{parse.unquote(name)}',
                    size_bytes=segment.length,
                )
            )

        # Try once and collect failed results to retry
        segment_results, retry_results = self._connection._wait_for_futures(
            segment_futures, raise_on_error=False
        )

        self._add_etag_to_manifest(segment_results, manifest)

        for result in retry_results:
            # Grab the FileSegment for the failed upload so we can retry
            name = self._object_name_from_url(result.url)
            segment = segments[name]
            segment.seek(0)
            # Async call to put - schedules execution and returns a future
            segment_future = self._connection._pool_executor.submit(
                self.put, name, headers=headers, data=segment
            )
            # TODO(mordred) Collect etags from results to add to this manifest
            # dict. Then sort the list of dicts by path.
            retry_futures.append(segment_future)

        # If any segments fail the second time, just throw the error
        segment_results, retry_results = self._connection._wait_for_futures(
            retry_futures, raise_on_error=True
        )

        self._add_etag_to_manifest(segment_results, manifest)

        try:
            if use_slo:
                return self._finish_large_object_slo(
                    endpoint, headers, manifest
                )
            else:
                return self._finish_large_object_dlo(endpoint, headers)
        except Exception:
            try:
                segment_prefix = endpoint.split('/')[-1]
                self.log.debug(
                    "Failed to upload large object manifest for %s. "
                    "Removing segment uploads.",
                    segment_prefix,
                )
                self._delete_autocreated_image_objects(
                    segment_prefix=segment_prefix
                )
            except Exception:
                self.log.exception(
                    "Failed to cleanup image objects for %s:", segment_prefix
                )
            raise

    def _finish_large_object_slo(self, endpoint, headers, manifest):
        # TODO(mordred) send an etag of the manifest, which is the md5sum
        # of the concatenation of the etags of the results
        headers = headers.copy()
        retries = 3
        while True:
            try:
                return exceptions.raise_from_response(
                    self.put(
                        endpoint,
                        params={'multipart-manifest': 'put'},
                        headers=headers,
                        data=json.dumps(manifest),
                    )
                )
            except Exception:
                retries -= 1
                if retries == 0:
                    raise

    def _finish_large_object_dlo(self, endpoint, headers):
        headers = headers.copy()
        headers['X-Object-Manifest'] = endpoint
        retries = 3
        while True:
            try:
                return exceptions.raise_from_response(
                    self.put(endpoint, headers=headers)
                )
            except Exception:
                retries -= 1
                if retries == 0:
                    raise

    def _upload_object(self, endpoint, filename, headers):
        with open(filename, 'rb') as dt:
            return self.put(endpoint, headers=headers, data=dt)

    def _get_file_segments(self, endpoint, filename, file_size, segment_size):
        # Use an ordered dict here so that testing can replicate things
        segments = collections.OrderedDict()
        for index, offset in enumerate(range(0, file_size, segment_size)):
            remaining = file_size - (index * segment_size)
            segment = _utils.FileSegment(
                filename,
                offset,
                segment_size if segment_size < remaining else remaining,
            )
            name = f'{endpoint}/{index:0>6}'
            segments[name] = segment
        return segments

    def get_object_segment_size(self, segment_size):
        """Get a segment size that will work given capabilities"""
        if segment_size is None:
            segment_size = DEFAULT_OBJECT_SEGMENT_SIZE
        min_segment_size = 0
        try:
            # caps = self.get_object_capabilities()
            caps = self.get_info()
        except (
            exceptions.NotFoundException,
            exceptions.PreconditionFailedException,
        ):
            server_max_file_size = DEFAULT_MAX_FILE_SIZE
            self._connection.log.info(
                "Swift capabilities not supported. "
                "Using default max file size."
            )
        except exceptions.SDKException:
            raise
        else:
            server_max_file_size = caps.swift.get('max_file_size', 0)
            min_segment_size = caps.slo.get('min_segment_size', 0)

        if segment_size > server_max_file_size:
            return server_max_file_size
        if segment_size < min_segment_size:
            return min_segment_size
        return segment_size

    def _object_name_from_url(self, url):
        '''Get container_name/object_name from the full URL called.

        Remove the Swift endpoint from the front of the URL, and remove
        the leaving / that will leave behind.'''
        endpoint = self.get_endpoint()
        object_name = url.replace(endpoint, '')
        if object_name.startswith('/'):
            object_name = object_name[1:]
        return object_name

    def _add_etag_to_manifest(self, segment_results, manifest):
        for result in segment_results:
            if 'Etag' not in result.headers:
                continue
            name = self._object_name_from_url(result.url)
            for entry in manifest:
                if entry['path'] == f'/{parse.unquote(name)}':
                    entry['etag'] = result.headers['Etag']

    def get_info(self):
        """Get infomation about the object-storage service

        The object-storage service publishes a set of capabilities that
        include metadata about maximum values and thresholds.
        """
        return self._get(_info.Info)

    def set_account_temp_url_key(self, key, secondary=False):
        """Set the temporary URL key for the account.

        :param key: Text of the key to use.
        :param bool secondary: Whether this should set the secondary key.
            (defaults to False)
        """
        account = self._get_resource(_account.Account, None)
        account.set_temp_url_key(self, key, secondary)

    def set_container_temp_url_key(self, container, key, secondary=False):
        """Set the temporary URL key for a container.

        :param container: The value can be the name of a container or a
            :class:`~openstack.object_store.v1.container.Container` instance.
        :param key: Text of the key to use.
        :param bool secondary: Whether this should set the secondary key.
            (defaults to False)
        """
        res = self._get_resource(_container.Container, container)
        res.set_temp_url_key(self, key, secondary)

    def get_temp_url_key(self, container=None):
        """Get the best temporary url key for a given container.

        Will first try to return Temp-URL-Key-2 then Temp-URL-Key for the
        container, and if neither exist, will attempt to return Temp-URL-Key-2
        then Temp-URL-Key for the account. If neither exist, will return None.

        :param container: The value can be the name of a container or a
            :class:`~openstack.object_store.v1.container.Container` instance.
        """
        temp_url_key = None
        if container:
            container_meta = self.get_container_metadata(container)
            temp_url_key = (
                container_meta.meta_temp_url_key_2
                or container_meta.meta_temp_url_key
            )
        if not temp_url_key:
            account_meta = self.get_account_metadata()
            temp_url_key = (
                account_meta.meta_temp_url_key_2
                or account_meta.meta_temp_url_key
            )
        if temp_url_key and not isinstance(temp_url_key, bytes):
            temp_url_key = temp_url_key.encode('utf8')
        return temp_url_key

    def _check_temp_url_key(self, container=None, temp_url_key=None):
        if temp_url_key:
            if not isinstance(temp_url_key, bytes):
                temp_url_key = temp_url_key.encode('utf8')
        else:
            temp_url_key = self.get_temp_url_key(container)
        if not temp_url_key:
            raise exceptions.SDKException(
                'temp_url_key was not given, nor was a temporary url key '
                'found for the account or the container.'
            )
        return temp_url_key

    def generate_form_signature(
        self,
        container,
        object_prefix,
        redirect_url,
        max_file_size,
        max_upload_count,
        timeout,
        temp_url_key=None,
    ):
        """Generate a signature for a FormPost upload.

        :param container: The value can be the name of a container or a
            :class:`~openstack.object_store.v1.container.Container` instance.
        :param object_prefix: Prefix to apply to limit all object names
            created using this signature.
        :param redirect_url: The URL to redirect the browser to after the
            uploads have completed.
        :param max_file_size: The maximum file size per file uploaded.
        :param max_upload_count: The maximum number of uploaded files allowed.
        :param timeout: The number of seconds from now to allow the form
            post to begin.
        :param temp_url_key: The X-Account-Meta-Temp-URL-Key for the account.
            Optional, if omitted, the key will be fetched from the container
            or the account.
        """
        max_file_size = int(max_file_size)
        if max_file_size < 1:
            raise exceptions.SDKException(
                'Please use a positive max_file_size value.'
            )
        max_upload_count = int(max_upload_count)
        if max_upload_count < 1:
            raise exceptions.SDKException(
                'Please use a positive max_upload_count value.'
            )
        if timeout < 1:
            raise exceptions.SDKException(
                'Please use a positive <timeout> value.'
            )
        expires = _get_expiration(timeout)

        temp_url_key = self._check_temp_url_key(
            container=container, temp_url_key=temp_url_key
        )

        res = self._get_resource(_container.Container, container)
        endpoint = parse.urlparse(self.get_endpoint())
        if isinstance(endpoint.path, bytes):
            # To keep mypy happy: the output type will be the same as the input
            # type
            path = endpoint.path.decode()
        else:
            path = endpoint.path
        path = '/'.join([path, res.name, object_prefix])

        data = f'{path}\n{redirect_url}\n{max_file_size}\n{max_upload_count}\n{expires}'  # noqa: E501
        sig = hmac.new(temp_url_key, data.encode(), sha1).hexdigest()

        return (expires, sig)

    def generate_temp_url(
        self,
        path,
        seconds,
        method,
        absolute=False,
        prefix=False,
        iso8601=False,
        ip_range=None,
        temp_url_key=None,
    ):
        """Generates a temporary URL that gives unauthenticated access to the
        Swift object.

        :param path: The full path to the Swift object or prefix if
            a prefix-based temporary URL should be generated. Example:
            /v1/AUTH_account/c/o or /v1/AUTH_account/c/prefix.
        :param seconds: time in seconds or ISO 8601 timestamp.
            If absolute is False and this is the string representation of an
            integer, then this specifies the amount of time in seconds for
            which the temporary URL will be valid.  If absolute is True then
            this specifies an absolute time at which the temporary URL will
            expire.
        :param method: A HTTP method, typically either GET or PUT, to allow
            for this temporary URL.
        :param absolute: if True then the seconds parameter is interpreted as a
            Unix timestamp, if seconds represents an integer.
        :param prefix: if True then a prefix-based temporary URL will be
            generated.
        :param iso8601: if True, a URL containing an ISO 8601 UTC timestamp
            instead of a UNIX timestamp will be created.
        :param ip_range: if a valid ip range, restricts the temporary URL to
            the range of ips.
        :param temp_url_key: The X-Account-Meta-Temp-URL-Key for the account.
            Optional, if omitted, the key will be fetched from the container or
            the account.
        :raises ValueError: if timestamp or path is not in valid format.
        :return: the path portion of a temporary URL
        """
        try:
            try:
                timestamp = float(seconds)
            except ValueError:
                formats = (
                    EXPIRES_ISO8601_FORMAT,
                    EXPIRES_ISO8601_FORMAT[:-1],
                    SHORT_EXPIRES_ISO8601_FORMAT,
                )
                for f in formats:
                    try:
                        t = time.strptime(seconds, f)
                    except ValueError:
                        continue

                    if f == EXPIRES_ISO8601_FORMAT:
                        timestamp = timegm(t)
                    else:
                        # Use local time if UTC designator is missing.
                        timestamp = int(time.mktime(t))

                    absolute = True
                    break
                else:
                    raise ValueError()
            else:
                if not timestamp.is_integer():
                    raise ValueError()
                timestamp = int(timestamp)
                if timestamp < 0:
                    raise ValueError()
        except ValueError:
            raise ValueError(
                'time must either be a whole number '
                'or in specific ISO 8601 format.'
            )

        if isinstance(path, bytes):
            try:
                path_for_body = path.decode('utf-8')
            except UnicodeDecodeError:
                raise ValueError('path must be representable as UTF-8')
        else:
            path_for_body = path

        parts = path_for_body.split('/', 4)
        if (
            len(parts) != 5
            or parts[0]
            or not all(parts[1 : (4 if prefix else 5)])
        ):
            if prefix:
                raise ValueError('path must at least contain /v1/a/c/')
            else:
                raise ValueError(
                    'path must be full path to an object e.g. /v1/a/c/o'
                )

        standard_methods = ['GET', 'PUT', 'HEAD', 'POST', 'DELETE']
        if method.upper() not in standard_methods:
            self.log.warning(
                'Non default HTTP method %s for tempurl '
                'specified, possibly an error',
                method.upper(),
            )

        expiration: float | int
        if not absolute:
            expiration = _get_expiration(timestamp)
        else:
            expiration = timestamp

        hmac_parts = [
            method.upper(),
            str(expiration),
            ('prefix:' if prefix else '') + path_for_body,
        ]

        if ip_range:
            if isinstance(ip_range, bytes):
                try:
                    ip_range = ip_range.decode('utf-8')
                except UnicodeDecodeError:
                    raise ValueError('ip_range must be representable as UTF-8')
            hmac_parts.insert(0, f"ip={ip_range}")

        hmac_body = '\n'.join(hmac_parts)

        temp_url_key = self._check_temp_url_key(temp_url_key=temp_url_key)

        sig = hmac.new(
            temp_url_key, hmac_body.encode('utf-8'), sha1
        ).hexdigest()

        if iso8601:
            exp = time.strftime(
                EXPIRES_ISO8601_FORMAT, time.gmtime(expiration)
            )
        else:
            exp = str(expiration)

        temp_url = f'{path_for_body}?temp_url_sig={sig}&temp_url_expires={exp}'

        if ip_range:
            temp_url += f'&temp_url_ip_range={ip_range}'

        if prefix:
            temp_url += f'&temp_url_prefix={parts[4]}'
        # Have return type match path from caller
        if isinstance(path, bytes):
            return temp_url.encode('utf-8')
        else:
            return temp_url

    def _delete_autocreated_image_objects(
        self, container=None, segment_prefix=None
    ):
        """Delete all objects autocreated for image uploads.

        This method should generally not be needed, as shade should clean up
        the objects it uses for object-based image creation. If something goes
        wrong and it is found that there are leaked objects, this method can be
        used to delete any objects that shade has created on the user's behalf
        in service of image uploads.

        :param str container: Name of the container. Defaults to 'images'.
        :param str segment_prefix: Prefix for the image segment names to
            delete. If not given, all image upload segments present are
            deleted.
        :returns: True if deletion was succesful, else False.
        """
        if container is None:
            container = self._connection._OBJECT_AUTOCREATE_CONTAINER
        # This method only makes sense on clouds that use tasks
        if not self._connection.image_api_use_tasks:
            return False

        deleted = False
        for obj in self.objects(container, prefix=segment_prefix):
            meta = self.get_object_metadata(obj).metadata
            if meta.get(self._connection._OBJECT_AUTOCREATE_KEY) == 'true':
                self.delete_object(obj, ignore_missing=True)
                deleted = True
        return deleted

    # ========== Utilities ==========

    def wait_for_status(
        self,
        res: resource.ResourceT,
        status: str,
        failures: list[str] | None = None,
        interval: int | float | None = 2,
        wait: int | None = None,
        attribute: str = 'status',
        callback: ty.Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for the resource to be in a particular status.

        :param session: The session to use for making this request.
        :param resource: The resource to wait on to reach the status. The
            resource must have a status attribute specified via ``attribute``.
        :param status: Desired status of the resource.
        :param failures: Statuses that would indicate the transition
            failed such as 'ERROR'. Defaults to ['ERROR'].
        :param interval: Number of seconds to wait between checks.
        :param wait: Maximum number of seconds to wait for transition.
            Set to ``None`` to wait forever.
        :param attribute: Name of the resource attribute that contains the
            status.
        :param callback: A callback function. This will be called with a single
            value, progress. This is API specific but is generally a percentage
            value from 0-100.

        :return: The updated resource.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if the
            transition to status failed to occur in ``wait`` seconds.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if the resource
            transitioned to one of the states in ``failures``.
        :raises: :class:`~AttributeError` if the resource does not have a
            ``status`` attribute
        """
        return resource.wait_for_status(
            self, res, status, failures, interval, wait, attribute, callback
        )

    def wait_for_delete(
        self,
        res: resource.ResourceT,
        interval: int = 2,
        wait: int = 120,
        callback: ty.Callable[[int], None] | None = None,
    ) -> resource.ResourceT:
        """Wait for a resource to be deleted.

        :param res: The resource to wait on to be deleted.
        :param interval: Number of seconds to wait before to consecutive
            checks.
        :param wait: Maximum number of seconds to wait before the change.
        :param callback: A callback function. This will be called with a single
            value, progress, which is a percentage value from 0-100.

        :returns: The resource is returned on success.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` if transition
            to delete failed to occur in the specified seconds.
        """
        return resource.wait_for_delete(self, res, interval, wait, callback)

    # ========== Project Cleanup ==========
    def _get_cleanup_dependencies(self):
        return {'object_store': {'before': []}}

    def _service_cleanup(
        self,
        dry_run=True,
        client_status_queue=None,
        identified_resources=None,
        filters=None,
        resource_evaluation_fn=None,
        skip_resources=None,
    ):
        if self.should_skip_resource_cleanup(
            "container", skip_resources
        ) or self.should_skip_resource_cleanup("object", skip_resources):
            return

        is_bulk_delete_supported = False
        bulk_delete_max_per_request = 1
        try:
            caps = self.get_info()
        except exceptions.SDKException:
            pass
        else:
            bulk_delete = caps.get("bulk_delete")
            if bulk_delete is not None:
                is_bulk_delete_supported = True
                bulk_delete_max_per_request = bulk_delete.get(
                    "max_deletes_per_request", 10000
                )

        elements = []
        for cont in self.containers():
            # Iterate over objects inside container
            objects_remaining = False
            for obj in self.objects(cont):
                need_delete = self._service_cleanup_del_res(
                    self.delete_object,
                    obj,
                    dry_run=True,
                    client_status_queue=client_status_queue,
                    identified_resources=identified_resources,
                    filters=filters,
                    resource_evaluation_fn=resource_evaluation_fn,
                )
                if need_delete:
                    if dry_run:
                        continue
                    elif is_bulk_delete_supported:
                        elements.append(f"{cont.name}/{obj.name}")
                        if len(elements) >= bulk_delete_max_per_request:
                            self._bulk_delete(elements)
                            elements.clear()
                    else:
                        self.delete_object(obj, cont)
                else:
                    objects_remaining = True

            if len(elements) > 0:
                self._bulk_delete(elements)
                elements.clear()

            # Eventually delete container itself
            if not objects_remaining:
                self._service_cleanup_del_res(
                    self.delete_container,
                    cont,
                    dry_run=dry_run,
                    client_status_queue=client_status_queue,
                    identified_resources=identified_resources,
                    filters=filters,
                    resource_evaluation_fn=resource_evaluation_fn,
                )

    def _bulk_delete(self, elements):
        data = "\n".join([parse.quote(x) for x in elements])
        self.delete(
            "?bulk-delete",
            data=data,
            headers={
                'Content-Type': 'text/plain',
                'Accept': 'application/json',
            },
        )
